/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "tserror.h"
#include "common.h"
#include "compact.h"
#ifndef TSCOMM
#define TSCOMM

#define LBUTTON 1
#define MBUTTON 2
#define RBUTTON 3

typedef struct _fameBuffer {
	int width;	/* fb dimensions */
	int height;
	int depth;	/* color depth of fb */
	int bpp;	/* how many bytes are used for each pixel */
	int stride;
	void* fb;
} TSframeBuffer;

typedef struct _cursor {
	int cursorSize; /*	If positive, has the size of the cursor in 
  						bytes (the source and the mask each are
						of size cursorSize)
						If -1, no cursor handling will be
						done by the client and the rest of the
						information in this message should be 
						disregarded */
	int cursorFG;	/* cursor colors */
	int cursorBG;
	int cursorWidth;/* cursor size in pixels */
	int cursorHeight;
	int xHot;
	int yHot;
	int x;			/* initial position x */
	int y;			/* initial position y */
	void* source;
	void* mask;
} TSCursor, *TSCursorPtr;

typedef struct thinc_rectRegion_s thinc_rectRegion;

/**/
extern int loopTimeoutInterval;

/* Cursor Functions */
int THINCHWCursorInit(ScreenPtr pScreen, TSCursorPtr pTSCursor);
int ThincCursorInit(ScreenPtr pScreen, TSCursorPtr pTSCursor);
Bool ThincInit(ScreenPtr pScreen, unsigned char *FBStart, 
			   int width, int height,  int depth, int bpp);

void thincShowCursor(ScreenPtr pScreen);
void thincHideCursor(ScreenPtr pScreen);
void thincSetCursorPosition(ScreenPtr pScreen, int x, int y);
void thincSetCursorColors(ScreenPtr pScreen, int bg, int fg);
void thincLoadCursorImage(ScreenPtr pScreen, unsigned char *bits);

PixmapPtr thincCreatePixmap(ScreenPtr pScreen,int width,int height, int depth);
Bool thincDestroyPixmap(PixmapPtr pPixmap);

typedef struct thincScreen_S thincScreenRec, *thincScreenPtr;

int thincBufferNewCmdReg(thincScreenPtr pThinc, WindowPtr pWindow, int type, 
	RegionPtr reg, ...);
int thincQueueCommand(thincScreenPtr pThinc, PixmapPtr pPix, int type,
	RegionPtr reg, ...);

int thincCopyPixmapQueue(thincScreenPtr pThinc, PixmapPtr srcpix, 
	PixmapPtr dstpix, int srcx, int srcy, int w, int h, int dstx, 
	int dsty);
int thincExecPixmapQueue(thincScreenPtr pThinc, WindowPtr dst, PixmapPtr pix, 
	int srcx, int srcy, int w, int h, int dstx, int dsty, 
	RegionPtr dstReg);

void thincMouseMove(int x, int y);
void thincMouseButton(int button, int isdown);
void thincKeyboardStroke(int vkey, int isdown);

void CursorWin32ToX(BYTE* source, BYTE* mask, int cursorSize);

void BlockDriverDrawing();
void UnblockDriverDrawing();
void UnblockDriverCopying();


#define AddEnabledDevice AddSocket
#define RemoveEnabledDevice RemoveSocket

//#define LESS_OVERWRITE

#endif

