/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdarg.h>
#include "TSDebug.h"
#include "windows.h"

int debugFlags = NO_DEBUG;
FILE* fileInstrument = NULL;
FILE* fileDebug = NULL;

void TSDebugStart(int flags)
{
	fileDebug = fopen("srvdebug.txt", "a");
	if (fileDebug == NULL)
	{
		debugFlags = NO_DEBUG;                    
	}
	else
	{
		debugFlags = flags;
		fprintf(fileDebug, "\n------------------------------------------------------\n\n");
	}                                        
}

void TSDebugEnd()
{
	if (fileDebug != NULL)
	{
		fflush(fileDebug);
		fclose(fileDebug);
		fileDebug = NULL;
	}
	debugFlags = NO_DEBUG;
}

void TSInstrumentStart()
{
	fileInstrument = fopen("srvinstru.txt", "a");
	if (fileInstrument != NULL)
	{
		fprintf(fileInstrument, "\n------------------------------------------------------\n\n");
	}
}

void TSInstrumentEnd()
{
	if (fileInstrument != NULL)
	{
		fflush(fileInstrument);
		fclose(fileInstrument);
		fileInstrument = NULL;
	}
}

int TSDebug(int flag, char* fmt, ...)
{
	va_list args;     
	int res;

#ifdef _DEBUG
	va_start(args, fmt);
	if (flag & debugFlags)
	{
		SYSTEMTIME st;
		GetSystemTime(&st);

		fprintf(fileDebug, "%.4d/%.2d/%.2d - %.2d:%.2d:%.2d:%.3d   ", st.wYear, st.wMonth, st.wDay, 
			st.wHour, st.wMinute, st.wSecond, st.wMilliseconds);
		res = vfprintf(fileDebug, fmt, args);
		fprintf(fileDebug, "\n");
	}
	va_end(args);
#endif
	return res;
}

int TSVDebug(int flag, char* fmt, va_list args)
{
	int res = 0;

#ifdef _DEBUG
	if (flag & debugFlags)
	{
		SYSTEMTIME st;
		GetSystemTime(&st);

		fprintf(fileDebug, "%.4d/%.2d/%.2d - %.2d:%.2d:%.2d:%.3d   ", st.wYear, st.wMonth, st.wDay, 
			st.wHour, st.wMinute, st.wSecond, st.wMilliseconds);
		res = vfprintf(fileDebug, fmt, args);
		fprintf(fileDebug, "\n");
	}
#endif
	return res;
}

int TSInstrument(char* fmt, ...)
{
	va_list args;     
	int res;

#ifdef _DEBUG
	va_start(args, fmt);
	if (fileInstrument != NULL)
	{
		FILETIME ft;
		SYSTEMTIME st;
		__int64 value;
		
		GetSystemTimeAsFileTime(&ft);
		FileTimeToSystemTime(&ft, &st);

		fprintf(fileInstrument, "%.4d/%.2d/%.2d - %.2d:%.2d:%.2d:%.3d   ", st.wYear, st.wMonth, st.wDay, 
			st.wHour, st.wMinute, st.wSecond, st.wMilliseconds);

		//memcpy(&value, &ft, sizeof(ft));
		//fprintf(fileInstrument, "%d,%d  ", ft.dwHighDateTime, ft.dwLowDateTime);
        res = vfprintf(fileInstrument, fmt, args);
		fprintf(fileInstrument, "\n");
	}
	va_end(args);
#endif
	return res;
}