// TTY Functions
#define XK_BackSpace         0xFF08
#define XK_Tab               0xFF09
#define XK_Clear             0xFF0B
#define XK_Return            0xFF0D
#define XK_Pause             0xFF13
#define XK_Scroll_Lock       0xFF14
#define XK_Escape            0xFF1B
#define XK_Delete            0xFFFF

// Cursor control & motion
#define XK_Home              0xFF50
#define XK_Left              0xFF51
#define XK_Up                0xFF52
#define XK_Right             0xFF53
#define XK_Down              0xFF54
#define XK_Page_Up           0xFF55
#define XK_Page_Down         0xFF56
#define XK_End               0xFF57

// Misc Functions
#define XK_Print             0xFF61
#define XK_Insert            0xFF63
#define XK_Undo              0xFF65
#define XK_Redo              0xFF66
#define XK_Find              0xFF68
#define XK_Cancel            0xFF69
#define XK_Help              0xFF6A
#define XK_Mode_switch       0xFF7E
#define XK_Num_Lock          0xFF7F

// Keypad functions
#define XK_KP_0              0xFFB0
#define XK_KP_1              0xFFB1
#define XK_KP_2              0xFFB2
#define XK_KP_3              0xFFB3
#define XK_KP_4              0xFFB4
#define XK_KP_5              0xFFB5
#define XK_KP_6              0xFFB6
#define XK_KP_7              0xFFB7
#define XK_KP_8              0xFFB8
#define XK_KP_9              0xFFB9

// function keys
#define XK_F1              0xFFBE
#define XK_F2              0xFFBF
#define XK_F3              0xFFC0
#define XK_F4              0xFFC1
#define XK_F5              0xFFC2
#define XK_F6              0xFFC3
#define XK_F7              0xFFC4
#define XK_F8              0xFFC5
#define XK_F9              0xFFC6
#define XK_F10             0xFFC7
#define XK_F11             0xFFC8
#define XK_F12             0xFFC9
#define XK_F13             0xFFCA
#define XK_F14             0xFFCB
#define XK_F15             0xFFCC
#define XK_F16             0xFFCD
#define XK_F17             0xFFCE
#define XK_F18             0xFFCF
#define XK_F19             0xFFD0
#define XK_F20             0xFFD1
#define XK_F21             0xFFD2
#define XK_F22             0xFFD3
#define XK_F23             0xFFD4
#define XK_F24             0xFFD5

// modifiers
#define XK_Shift_L         0xFFE1   /* Left shift */
#define XK_Shift_R         0xFFE2   /* Right shift */
#define XK_Control_L       0xFFE3   /* Left control */
#define XK_Control_R       0xFFE4   /* Right control */
#define XK_Caps_Lock       0xFFE5   /* Caps lock */
#define XK_Shift_Lock      0xFFE6   /* Shift lock */

#define XK_Meta_L          0xFFE7   /* Left meta */
#define XK_Meta_R          0xFFE8   /* Right meta */
#define XK_Alt_L           0xFFE9   /* Left alt */
#define XK_Alt_R           0xFFEA   /* Right alt */

#define XK_Semi_Colon		0x003D
#define VK_SEMICOLON		0x00CD

#define XK_Colon			0x003B
#define VK_COLON			0x00BA

#define XK_Single_Quote		0x0027
#define VK_SINGLE_QUOTE		0x00DE

#define XK_Back_Tick		0x0060
#define VK_BACK_TICK		0X00C0

#define XK_Back_slash		0x005C
#define VK_BACK_SLASH		0x00DC

#define XK_Period			0x002E
#define	VK_PERIOD			0x00BE

#define XK_Slash			0x002F
#define VK_SLASH			0x00BF

#define XK_Comma			0x002C
#define VK_COMMA			0x00BC

#define XK_RBracket			0x005B
#define VK_RBRACKET			0x00DB

#define XK_LBracket			0x005D
#define VK_LBRACKET			0x00DD

#define XK_Plus				0x003D
#define VK_PLUS				0x00BB

#define XK_Minus			0x002D
#define VK_MINUS			0x00BD

#define XKEY_SIZE 98

const int KeyMap[][2] = {
		VK_BACK,		XK_BackSpace,
		VK_TAB,			XK_Tab,
		VK_CLEAR,		XK_Clear,
		VK_RETURN,		XK_Return,
		VK_PAUSE,		XK_Pause,
		VK_SCROLL,		XK_Scroll_Lock,
		VK_ESCAPE,		XK_Escape,
		VK_DELETE,		XK_Delete,

		VK_HOME,		XK_Home,
		VK_UP,			XK_Up,
		VK_LEFT,		XK_Left,
		VK_RIGHT,		XK_Right,
		VK_DOWN,		XK_Down, 
		VK_PRIOR,		XK_Page_Up,
		VK_NEXT,		XK_Page_Down,
		VK_END,			XK_End,

		VK_PRINT,		XK_Print,
		VK_INSERT,		XK_Insert,
		//VK_UNDO,		XK_Undo,
		//VK_REDO,		XK_Redo,
		//VK_FIND,		XK_Find,
		//VK_CANCEL,	XK_Cancel,
		//VK_HELP,		XK_Help,  
		//VK_MODE_SWITCH,	XK_Mode_switch,
		VK_NUMLOCK,		XK_Num_Lock,

		// Keypad functions
		VK_NUMPAD0,		XK_KP_0,
		VK_NUMPAD1,		XK_KP_1,
		VK_NUMPAD2,		XK_KP_2,
		VK_NUMPAD3,		XK_KP_3,
		VK_NUMPAD4,		XK_KP_4,
		VK_NUMPAD5,		XK_KP_5,
		VK_NUMPAD6,		XK_KP_6,
		VK_NUMPAD7,		XK_KP_7,
		VK_NUMPAD8,		XK_KP_8,
		VK_NUMPAD9,		XK_KP_9,

		VK_F1,			XK_F1,
		VK_F2,			XK_F2,	
		VK_F3,			XK_F3,
		VK_F4,			XK_F4,
		VK_F5,			XK_F5,
		VK_F6,			XK_F6,
		VK_F7,			XK_F7,
		VK_F8,			XK_F8,
		VK_F9,			XK_F9,
		VK_F10,			XK_F10,
		VK_F11,			XK_F11,
		VK_F12,			XK_F12,
		VK_F13,			XK_F13,
		VK_F14,			XK_F14,
		VK_F15,			XK_F15,
		VK_F16,			XK_F16,
		VK_F17,			XK_F17,
		VK_F18,			XK_F18,
		VK_F19,			XK_F19,
		VK_F20,			XK_F20,
		VK_F21,			XK_F21,
		VK_F22,			XK_F22,
		VK_F23,			XK_F23,
		VK_F24,			XK_F24,

		VK_LSHIFT,		XK_Shift_L,
		VK_RSHIFT,		XK_Shift_R,
		VK_LCONTROL,	XK_Control_L,
		VK_RCONTROL,	XK_Control_R,
		VK_CAPITAL,		XK_Caps_Lock,
		VK_SHIFT,		XK_Shift_L,	
		VK_CONTROL,		XK_Control_L,	

		//VK_META_L,	XK_Meta_L,   
		//VK_META_R,	XK_Meta_R,
		VK_LMENU,		XK_Alt_L,   
		VK_RMENU,		XK_Alt_R,   

		//VK_SEMICOLON,	XK_Semi_Colon,
		//VK_BACK_SLASH,	XK_Back_slash,
		VK_COLON,		XK_Colon,
		VK_PERIOD,		XK_Period,
		VK_SLASH,		XK_Slash,
		VK_COMMA,		XK_Comma,
		VK_LBRACKET,	XK_LBracket,
		VK_RBRACKET,	XK_RBracket,
		VK_PLUS,		XK_Plus,
		VK_MINUS,		XK_Minus,
		VK_SINGLE_QUOTE,XK_Single_Quote,
		VK_BACK_TICK,	XK_Back_Tick,

		// {'A' to 'Z'} --> {'a' to 'z'}
		65,				97,
		66,				98,
		67,				99,
		68,				100,
		69,				101,
		70,				102,
		71,				103,
		72,				104,
		73,				105,
		74,				106,
		75,				107,
		76,				108,
		77,				109,
		78,				110,
		79,				111,
		80,				112,
		81,				113,
		82,				114,
		83,				115,
		84,				116,
		85,				117,
		86,				118,
		87,				119,
		88,				120,
		89,				121,
		90,				122,
};	
