/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 *
 * Bit insertion and extraction utilities
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 * With some inspiration from routines written by
 * Dick Hogaboom
 * http://www.embedded.com/1999/9907/9907feat2.htm
 */

#include "thinc.h"
#ifndef T_WIN32
#include <netinet/in.h>
#endif
/*
 * Insert bit_len bits from value into cptr starting at bit_offset. bit_offset
 * counts from the right (least significant bit is offset 0). It won't erase
 * what cptr had before.
 */
void t_bit_insert(unsigned char *cptr, unsigned long bit_offset,
	unsigned long bit_len, long value)
{
	/* if the value you're inserting has bits set outside of the range
	 * you want to insert, then you need to uncomment this */
#if 0
	unsigned long bits = t_bit_extract((unsigned char*)&value, 0, bit_len);
#else
	unsigned long bits = value;
#endif
	unsigned long l = *(long*)cptr;
	bits <<= bit_offset;
	l |= bits;
	*(long*)cptr = l;
}

/*
 * returns bit_len bits in cptr starting at bit_offset. note
 * that bit_offset counts from the right (least significant bit is offset 0).
 * The bits will be returned in the least significant portion of the return 
 * value
 */
unsigned long t_bit_extract(const unsigned char *cptr, 
	unsigned long bit_offset, unsigned long bit_len)
{
	unsigned long left = T_PIXELS_PER_LONG - (bit_offset + bit_len);
	unsigned long right = left + bit_offset;
	unsigned long l = *(unsigned long*)cptr;

	return (l << left) >> right;
}

/* this is useful for debugging purposes */
void dec_to_bin(unsigned long n)
{
	int i=0 , b;
	char bin[33], *curr=bin;
	
#if 0   /* The X way */
	for(i=0;i<32;i++) {
#else   /* The "intuitive" way */
	for(i=31; i>=0; i--) {
#endif
			b = (((1<<i)&n)>0) ? 1 : 0 ;
			sprintf(curr++, "%d",b);
	}
	bin[32] = 0;
	thincDbg(T_DBG_DRAW, ("%s: %s\n", __func__, bin));
}
