/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * 
 * ThincKeyboard functions
 * 
 * Authors:
 * Ricardo Baratto
 *
 * Network Computing Lab - Columbia University
 */

/* Heavily based on rfbkeyb from xf4vnc (xf4vnc.sourceforge.net) */

/* set tabstop = 8 */
#ifdef T_WIN32
#include "thinc.h"
#include "keymap.h"
#else /* T_WIN32 */
#include <sys/time.h>
#include <time.h>
#include <netinet/in.h>
#include "thinc.h"

#include "xf86.h"
#include "xf86Xinput.h"
#include "mi.h"
#include "mipointer.h"
#endif /*T_WIN32 */

#ifndef STATIC

#ifdef XKB
#include "X11/extensions/XKB.h"
#include "X11/extensions/XKBstr.h"
#include "X11/extensions/XKBsrv.h"

static char * xkbkeymap;
static char * xkbkeycodes;
static char * xkbtypes;
static char * xkbcompat;
static char * xkbsymbols;
static char * xkbgeometry;
static Bool xkbcomponents_specified;
static char * xkbrules;
static char * xkbmodel;
static char * xkblayout;
static char * xkbvariant;
static char * xkboptions;
#endif

/* event handling functions */
static void thincInsertKeybEvent(thinc_evKeyb ev, unsigned long time);
#ifndef T_WIN32
static InputInfoPtr thincKeybInit(InputDriverPtr drv, IDevPtr dev, int flags);
static void thincKeybUninit(InputDriverPtr drv, InputInfoPtr info, int flags);
static int thincKbdDeviceInit(DeviceIntPtr dev, KeySymsPtr keySyms,
	CARD8 *modMap);
static int thincKeybControl(DeviceIntPtr device, int onoff);
#endif /* T_WIN32*/
#endif

#ifndef T_WIN32
static DeviceIntPtr thincKbdDevice = NULL;

#include "thincKeyb.h"

#ifndef STATIC

static InputDriverRec ThincKeyb = {
	THINC_VERSION_MAJOR,    /* version */
        "thinckeyb",   	/* name */
        NULL,          	/* identify */
        thincKeybInit, 	/* preinit */
        thincKeybUninit,	/* uninit */
        NULL,           /* module */
        0               /* ref count */
};

static const char *OPT_DEFAULTS[] = {
        NULL
};

#endif

#ifdef XINPUT
#include "xf86Xinput.h"
#define enqueue(e)	xf86eqEnqueue(e)
#else
#define enqueue(e)	mieqEnqueue(e)
#endif
#endif /* T_WIN32*/

int thincHandleKeybEv(thincClientPtr cl)
{
	t_dbg_name(func)
	int res;
	thinc_evKeyb k;
	unsigned long time;

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	res = thincReadAnyMsg(cl, &k, T_EV_KEYB_SIZE);
	if(res == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s Couldn't read keyb event "
			"message\n", func));
		return res;
	} else if(res == T_CLOSED) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Client closed connection "
			"while reading keyb event message\n", func));
		return res;
	}

	k.key = ntohl(k.key);

	gettime_ms(time);

	thincDbg(T_DBG_INPUT, ("%s: Got keyboard event for key %d, pressed? "
		"%d\n",	func, k.key, k.down));

	/* inject the event */
	thincInsertKeybEvent(k, time);

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));

	return T_OK;
}

static void thincInsertKeybEvent(thinc_evKeyb ev, unsigned long time)
{
	t_dbg_name(func)
#ifdef T_WIN32
	int vkey = -1;
	int i;
	for (i = 0; i < XKEY_SIZE; i++)
	{
		if (KeyMap[i][1] == ev.key) 
		{
			vkey = KeyMap[i][0];
			break;
		}
	}
	if (vkey == -1) 
	{
		/* No vkey found*/
		vkey = ev.key;
	}
	thincKeyboardStroke(vkey, ev.down);
	
#else /* T_WIN32 */
	KeySymsPtr keySyms;
	int i, keycode;
	KeySym key = ev.key;
	xEvent e;

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	if(!thincKbdDevice) {
		thincError("ERROR: No keyboard device found, can't add "
			"keyboard event\n");
		return;
	}

	keySyms = &thincKbdDevice->key->curKeySyms;

	if(ev.down == T_DOWN)
		e.u.u.type = KeyPress;
	else
		e.u.u.type = KeyRelease;

	/****** find the keycode - from VNC:kbdptr.c:KbdAddEvent *****/
	keycode = 0;

	thincDbg(T_DBG_INPUT, ("%s: looking for keysym %#lx\n", func, key));
	
	/* first, see if it's a key we know about */
#if 0
	for(i=0; i< NUM_PREDEF_KEYS; i++) {
		if(key == thincMap[i]) {
			keycode = MIN_KEYCODE + i / GLYPHS_PER_KEY;
			break;
		}
	}
#else
	for(i=0; i< NUM_PREDEF_KEYS; i++) {
		if(key == thincMap[i*GLYPHS_PER_KEY]) {
			keycode = MIN_KEYCODE + i;
			break;
		}
	}
#endif

	thincDbg(T_DBG_INPUT, ("%s: found keycode %#x\n", func, keycode));

	e.u.keyButtonPointer.time = time;
	e.u.u.detail = keycode;
	enqueue(&e);
#endif /* T_WIN32 */
	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
}
#ifndef T_WIN32
#ifndef STATIC
/* TODO */
static void thincSendBell()
{
	/*
	thincLog("Keyboard Bell\n");
	*/
}

static InputInfoPtr thincKeybInit(InputDriverPtr drv, IDevPtr dev, int flags)
{
	t_dbg_name(func)
        /* seems to be an structure with information about the input
         * device. Steps: Allocate, fill in, process options (none for us)
         * and return so it gets registered */
	InputInfoPtr info;
	char *curr;
	int from;

	thincDbg(T_DBG_INPUT, ("%s\n", func));
	
	if(!(info = xf86AllocateInput(drv, 0)))
		return NULL;

	/* fill the structure */
	info->name = dev->identifier;
	info->type_name = "thincKeyb";
	info->flags = XI86_KEYBOARD_CAPABLE;
	info->device_control = thincKeybControl;
	info->read_input = NULL;
	info->motion_history_proc = NULL;
	info->history_size = 0;
	info->control_proc = NULL;
	info->close_proc = NULL;
	info->switch_mode = NULL;
	info->conversion_proc = NULL;
	info->reverse_conversion_proc = NULL;
	info->fd = -1;
	info->dev = NULL;
	info->private_flags = 0;
	info->always_core_feedback = 0;
	info->conf_idev = dev;

        /* collect options and process the common ones. For now we
         * don't have any options. Leaving here just in case */
        xf86CollectInputOptions(info, OPT_DEFAULTS, NULL);
        xf86ProcessCommonOptions(info, info->options);

	/* Mark device as configured */
	info->flags |= XI86_CONFIGURED;

	/* XKB Stuff. From xf4vnc */
#ifdef XKB
	from = X_DEFAULT;
	if(noXkbExtension)
		from = X_CMDLINE;
	else if(xf86FindOption(dev->commonOptions, "XkbDisable")) {
		noXkbExtension = xf86SetBoolOption(dev->commonOptions,
			"XkbDisable", FALSE);
		from = X_CONFIG;
	}
	if(noXkbExtension)
		xf86Msg(from, "XKB disabled\n");

#define NULL_IF_EMPTY(s) (curr[0] ? s: (free(curr), (char*)NULL))

	if(!noXkbExtension && !XkbInitialMap) {
		if((curr = xf86SetStrOption(dev->commonOptions, "XkbKeymap", 
			NULL))) {
			xkbkeymap = NULL_IF_EMPTY(curr);
			xf86Msg(X_CONFIG, "XKB: Keymap: %s\n", xkbkeymap);
		} else {
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbCompat", NULL))) {
				xkbcompat = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: compat %s\n",xkbcompat);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbTypes", NULL))) {
				xkbtypes = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: types %s\n", xkbtypes);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbKeycodes", NULL))) {
				xkbkeycodes = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: keycodes %s\n",
					xkbkeycodes);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbGeometry", NULL))) {
				xkbgeometry = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: geometry %s\n",
					xkbgeometry);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbSymbols", NULL))) {
				xkbsymbols = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: symbols %s\n",
					xkbsymbols);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbRules", NULL))) {
				xkbrules = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: rules %s\n",
					xkbrules);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbModel", NULL))) {
				xkbmodel = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: model %s\n",
					xkbmodel);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbLayout", NULL))) {
				xkblayout = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: layout %s\n",
					xkblayout);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbVariant", NULL))) {
				xkbvariant = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: variant %s\n",
					xkbvariant);
			}
			if((curr = xf86SetStrOption(dev->commonOptions,
				"XkbOptions", NULL))) {
				xkboptions = NULL_IF_EMPTY(curr);
				xkbcomponents_specified = TRUE;
				xf86Msg(X_CONFIG, "XKB: options %s\n",
					xkboptions);
			}
		}
	}
#undef NULL_IF_EMPTY
#endif	/* XKB */

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));

	return info;
}

static void thincKeybUninit(InputDriverPtr drv, InputInfoPtr info, int flags)
{
	thincKeybControl(info->dev, DEVICE_OFF);
}

static int thincKeybControl(DeviceIntPtr device, int onoff)
{
	t_dbg_name(func)
	KeySymsRec keySyms;
	CARD8 modMap[MAP_LENGTH];
	DevicePtr dev = (DevicePtr)device;
	int res;

	thincDbg(T_DBG_INPUT, ("%s: onoff %d\n", func, onoff));

	switch(onoff) {
		case DEVICE_INIT:
			res = thincKbdDeviceInit(device, &keySyms, modMap);
			if(res != Success) {
				thincDbg(T_DBG_INPUT, ("%s: finished\n", func));
				return res;
			}
#ifdef XKB
			if(noXkbExtension) {
#endif
			InitKeyboardDeviceStruct(dev, &keySyms, modMap,
				(BellProcPtr)thincSendBell,
				(KbdCtrlProcPtr)NoopDDA);
#ifdef XKB
			} else {
				XkbComponentNamesRec names;
				if(XkbInitialMap) {
					if((xkbkeymap = strchr(XkbInitialMap,
						'/')) != NULL)
						xkbkeymap++;
					else
						xkbkeymap = XkbInitialMap;
				}
				if(xkbkeymap) {
					names.keymap = xkbkeymap;
					names.keycodes = NULL;
					names.types = NULL;
					names.compat = NULL;
					names.symbols = NULL;
					names.geometry = NULL;
				} else {
					names.keymap = NULL;
					names.keycodes = xkbkeycodes;
					names.types = xkbtypes;
					names.compat = xkbcompat;
					names.symbols = xkbsymbols;
					names.geometry = xkbgeometry;
				}
				if((xkbkeymap || xkbcomponents_specified) &&
				   (xkbmodel == NULL || xkblayout == NULL)) {
					xkbrules = NULL;
				}
				XkbInitKeyboardDeviceStruct(device,
					&names, &keySyms, modMap,
					(BellProcPtr)thincSendBell,
					(KbdCtrlProcPtr)NoopDDA);
			}
#endif
			break;
		case DEVICE_ON:
			dev->on = TRUE;
			break;
		case DEVICE_OFF:
			dev->on = FALSE;
			break;
		case DEVICE_CLOSE:
			break;
	}

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));

	return Success;
}

/* fills in the keysyms and keymap */
static int thincKbdDeviceInit(DeviceIntPtr dev, KeySymsPtr keySyms,
	CARD8 *modMap)
{
	int i;
	t_dbg_name(func)

	thincDbg(T_DBG_INPUT, ("%s\n", func));

	thincKbdDevice = dev;

	for(i=0; i<MAP_LENGTH; i++)
		modMap[i] = NoSymbol;

	modMap[CONTROL_L_CODE] = ControlMask;
	modMap[CONTROL_R_CODE] = ControlMask;
	modMap[SHIFT_L_CODE] = ShiftMask;
	modMap[SHIFT_R_CODE] = ShiftMask;
	modMap[META_L_CODE] = Mod1Mask;
	modMap[META_R_CODE] = Mod1Mask;
	modMap[ALT_L_CODE] = Mod1Mask;
	modMap[ALT_R_CODE] = Mod1Mask;

	keySyms->minKeyCode = MIN_KEYCODE;
	keySyms->maxKeyCode = MAX_KEYCODE;
	keySyms->mapWidth = GLYPHS_PER_KEY;

	if(!(keySyms->map = (KeySym *)malloc(sizeof(KeySym)*MAP_LENGTH *
		GLYPHS_PER_KEY))) {
			thincError("KbdDeviceInit: failed allocate keyboard "
				"map\n");
			return BadAlloc;
	}
	
	for(i=0; i<MAP_LENGTH *GLYPHS_PER_KEY; i++)
		keySyms->map[i] = NoSymbol;

	for(i=0; i< NUM_PREDEF_KEYS * GLYPHS_PER_KEY; i++)
		keySyms->map[i] = thincMap[i];

	thincDbg(T_DBG_INPUT, ("%s: finished\n", func));

	return Success;
}

/***** dynamic loader stuff ******/
#ifdef XFree86LOADER
void thincInitKeyboard(void)
{
	xf86AddInputDriver(&ThincKeyb, NULL, 0);
}

static void thincKeybUnplug(pointer p)
{
}

static pointer thincKeybPlug(pointer module, pointer options, int *errmaj,
	int *errmin)
{
	xf86AddInputDriver(&ThincKeyb, module, 0);
	
	return module;
}

static XF86ModuleVersionInfo thincKeybVersionRec = {
	"thinckeyb",
	MODULEVENDORSTRING,
	MODINFOSTRING1,
	MODINFOSTRING2,
	XF86_VERSION_CURRENT,
	THINC_VERSION_MAJOR, THINC_VERSION_MINOR, THINC_PATCHLEVEL,
	ABI_CLASS_XINPUT,
	ABI_XINPUT_VERSION,
	MOD_CLASS_XINPUT,
	{0, 0, 0, 0}
};

XF86ModuleData thincKeybModuleData = {
	&thincKeybVersionRec,
	thincKeybPlug,
	thincKeybUnplug
};
#else
void thincInitKeyboard(void)
{
}
#endif

#endif
#endif /* T_WIN32*/
