/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 *
 * Pixmap functions
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 */

/* set tabstop = 8 */

#include "thinc.h"

/* enable pixmap write to file function. you must also #define 
 * T_ENABLE_RAWRITE_FUNC in thincCommandRaw.c */
#if 0
#define T_ENABLE_PIXWRITE_FUNC
#endif

#define T_CMD_FIRST_ID	(1<<31)

PixmapPtr thincCreatePixmap(ScreenPtr pScreen,int width,int height, int depth)
{
	t_dbg_name(func)
	thincScreenPtr pThinc = THINCPtr(pScreen);
	PixmapPtr pPixmap;

	thincDbg(T_DBG_CMD, ("%s: Pixmap dimensions: width %d, height %d, "
		"depth %d\n", func, width, height, depth));

#ifdef T_WIN32
	pPixmap = malloc(sizeof(PixmapStr));
	if (pPixmap) 
	{
		pPixmap->width = width;
		pPixmap->height = height;
		pPixmap->bpp = depth;	/* FIXME: Confirmed??*/
		pPixmap->pThincPixmap = malloc(sizeof(thincPixmapRec));
		pPixmap->pThinc = pThinc;
		if (pPixmap->pThincPixmap == NULL)
		{
			free(pPixmap);
			pPixmap = NULL;
		}
	}
#else /* T_WIN32 */
	thinc_unwrap(pThinc, pScreen, CreatePixmap);
	pPixmap = (*pScreen->CreatePixmap)(pScreen,width, height, depth);
	thinc_wrap(pThinc, pScreen, CreatePixmap, thincCreatePixmap);
#endif /* T_WIN32 */
	if(pPixmap) {
		thincDbg(T_DBG_CMD, ("%s: Created pixmap with SN %ld, "
			"refcount %d\n", func, pPixmap->drawable.serialNumber, 
			pPixmap->refcnt));
	} else {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create pixmap\n",
			func));
	}

	/* initialize our devPrivate. We don't allocate memory for the
	 * cmdlist until we actually need to, to avoid pixmaps which won't
	 * have any commands sent to them */
	thincInitPixmapPriv(pPixmap);

	thincDbg(T_DBG_CMD, ("%s: finished. Pixmap id %#lx. Data %p\n", 
		func, t_pixID(pPixmap), pPixmap->devPrivate.ptr));

	return pPixmap;
}

#ifdef T_USEMITSHM
/* fbShmCreatePixmap with a thinc twist. Assumes we're using offscreen
 * pixmaps. If not, there's no point in calling this function, just use
 * mi's */
PixmapPtr thincShmCreatePixmap(ScreenPtr pScreen, int width,
	int height, int depth, char *addr)
{
	t_dbg_name(func)
	register PixmapPtr pPixmap;
	int bpp;
	RegionRec r;
	thincScreenPtr pThinc = THINCPtr(pScreen);

	thincDbg(T_DBG_CMD, ("%s: Creating SHM pixmap of size %dx%d and "
		"depth %d. data %p\n", func, width, height, depth, addr));

	pPixmap = thincCreatePixmap(pScreen, 0, 0, pScreen->rootDepth);
	if(!pPixmap) {
		thincDbg(T_DBG_ERROR, ("%s: Couldn't create SHM pixmap\n",
			func));
		return NullPixmap;
	}

	if(depth == 24)
		bpp = 32;
	else
		bpp = depth;

	if(!(thincModifyPixmapHeader(pPixmap, width, height, depth,
		bpp, PixmapBytePad(width, depth), (pointer)addr)) ) {
		thincDestroyPixmap(pPixmap);

		thincDbg(T_DBG_ERROR, ("%s: Couldn't create SHM Pixmap\n",
			func));
		return NullPixmap;
	}

	/* Since the X app can modify the pixmap's data directly,
	 * we have no way to know what's being drawn there, so this
	 * is the only way :( */
	coords_to_reg(pScreen, &r, 0, 0, width, height);
	thincQueueCommand(pThinc, pPixmap, T_FB_RUPDATE, &r, 0, NULL);
	REGION_UNINIT(pScreen, &r);
	thincDbg(T_DBG_CMD, ("%s: finished. Pixmap ID %#lx\n", func,
		t_pixID(pPixmap)));

	return pPixmap;
}
#endif

#ifndef T_WIN32
Bool thincModifyPixmapHeader(PixmapPtr pPixmap, int width, int height,
	int depth, int bitsPerPixel, int devKind, pointer pPixData)
{
	t_dbg_name(func)
	Bool res;
	ScreenPtr pScreen = pPixmap->drawable.pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);

	thincDbg(T_DBG_CMD, ("%s: modifying pixmap with SN %ld, id %#lx. "
		"refcount %d\n", func, pPixmap->drawable.serialNumber, 
		t_pixID(pPixmap), pPixmap->refcnt));

	thincDbg(T_DBG_CMD, ("%s: width %d height %d depth %d "
		"bitsPerPixel %d, devKind %d, pPixData %p\n", func, 
		width, height, depth, bitsPerPixel, devKind, pPixData));

	thinc_unwrap(pThinc, pScreen, ModifyPixmapHeader);
	res = (*pScreen->ModifyPixmapHeader)(pPixmap, width, height,
		depth, bitsPerPixel, devKind, pPixData);
	thinc_wrap(pThinc, pScreen, ModifyPixmapHeader, 
		thincModifyPixmapHeader);

	thincDbg(T_DBG_CMD, ("%s: finished, data %p\n", func, 
		pPixmap->devPrivate.ptr));
	return res;
}
#endif /* T_WIN32 */

Bool thincDestroyPixmap(PixmapPtr pPixmap)
{
	t_dbg_name(func)
#ifndef T_WIN32
	ScreenPtr pScreen = pPixmap->drawable.pScreen;
	thincScreenPtr pThinc = THINCPtr(pScreen);
#endif /* T_WIN32 */
	Bool res;

	thincDbg(T_DBG_CMD, ("%s: Destroying pixmap, id %#lx, "
		"refcnt %d\n", func, t_pixID(pPixmap), pPixmap->refcnt));
#ifndef T_WIN32
	if(pPixmap->refcnt == 1)
#endif /* T_WIN32 */
		thincUninitPixmapPriv(pPixmap);

#ifdef T_WIN32
	if (pPixmap != NULL)
	{
		if (pPixmap->pThincPixmap != NULL)
			free(pPixmap->pThincPixmap);
		free(pPixmap);
		res = TRUE;
	}
	else
		res = FALSE;
#else /* T_WIN32 */
	thinc_unwrap(pThinc, pScreen, DestroyPixmap);
	res = (*pScreen->DestroyPixmap)(pPixmap);
	thinc_wrap(pThinc, pScreen, DestroyPixmap, thincDestroyPixmap);
#endif /* T_WIN32 */
	thincDbg(T_DBG_CMD, ("%s: finished\n", func));

	return res;
}

/* initializes our devPrivate structure to default values. No memory is
 * allocated for the command list. We do this the first time a command
 * needs to be queued into the list */
void thincInitPixmapPriv(PixmapPtr pPix)
{
	t_dbg_name(func)
	static ulong id = T_CMD_FIRST_ID;
	thincPixmapPtr pPriv = THINCPIXPtr(pPix);

	thincDbg(T_DBG_CMD, ("%s: initializing pixmap private\n", func));

	pPriv->id = id--;
	t_list_init(&pPriv->list);

#ifdef T_USERENDER
	pPriv->i_got_it = 0;
#endif

	thincDbg(T_DBG_CMD, ("%s: finished. id %#lx\n", func, pPriv->id));
}

void thincUninitPixmapPriv(PixmapPtr pPix)
{
	thincPixmapPtr pPriv = THINCPIXPtr(pPix);
#ifdef T_WIN32
	thincScreenPtr pThinc = (thincScreenPtr)pPix->pThinc;
#else
	thincScreenPtr pThinc = THINCPtr(pPix->drawable.pScreen);
#endif

	stats_dbg(("Destroy:Pixmap[%#lx]: len %u max_len %u\n", 
		t_pixID(pPix), t_list_len(&pPriv->list), 
		t_list_maxlen(&pPriv->list)));
	t_list_len_init(&pPriv->list);

	if(!t_list_is_empty(&pPriv->list)) {
		/* don't use destroy since there's nothing to free() */
		thincEmptyCmdListBuffer(pThinc, &pPriv->list);
	}
}

void thincEmptyPixmapQueue(PixmapPtr pPix)
{
	thincPixmapPtr pPriv = THINCPIXPtr(pPix);
#ifdef T_WIN32	
	thincScreenPtr pThinc = pPix->pThinc;
#else
	thincScreenPtr pThinc = THINCPtr(pPix->drawable.pScreen);
#endif 

	if(!t_list_is_empty(&pPriv->list)) {
		thincEmptyCmdListBuffer(pThinc, &pPriv->list);
		/* at some point this queue was empty */
		t_list_len_set(&pPriv->list, 0);
	}
}


#ifdef T_ENABLE_PIXWRITE_FUNC
extern void thincRawWriteToFile(char *data, unsigned int w,
        unsigned int h, unsigned int Bpp, unsigned int pitch);

void thincPixmapWriteToFile(PixmapPtr pPix)
{
	thincRawWriteToFile(thinc_getPixmap(pPix), pPix->drawable.width,
		pPix->drawable.height, 
		pPix->drawable.bitsPerPixel >> 3, /* hack */
		pPix->devKind);
}
#endif
