/*
 *
 * Header for thinc's data cache
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 */

/* set tabstop = 8 */

/* Implemented using a hash table (see thincHash.c) */

#include "thincCommon.h"

/**** hash ****/
#define thinc_hashsize(n) ((unsigned long)1<<(n))
#define thinc_hashmask(n) (thinc_hashsize(n)-1)
unsigned long thinc_hash(register unsigned char *k, 
	register unsigned long length, register unsigned long initval);

/* an entry in the cache */
typedef struct thinc_cache_el {
	unsigned long int hash;
	unsigned int flags;
	void *priv;	/* assumed to be shallow for simplicity */
} thinc_cache_el;

/* flags */
#define T_EMPTY		0x00
#define T_ACTIVE	0x01

/* the cache object */
typedef struct thinc_cache_s {
	thinc_cache_el *cache;
	unsigned int	idx_len;  /* how many bits to use as index in
				     hash table */
	unsigned int	total;
	unsigned int	active;	  /* how many are active */
	unsigned int	collisions;
} thinc_cache;

#define thinc_cache_size(cache) thinc_hashsize((cache)->idx_len)

thinc_cache *thinc_cache_init(int);
void thinc_cache_destroy(thinc_cache*);
void thinc_cache_empty(thinc_cache*);
void thinc_cache_destroy_cb(thinc_cache* cache, 
	void (*callback)(void *, void *), void *ptr);
void thinc_cache_empty_cb(thinc_cache* cache, void (*callback)(void *, void *),
	void *ptr);
int thinc_cache_get_size(thinc_cache *cache);
int thinc_cache_get_active(thinc_cache *cache);
int thinc_cache_get_collisions(thinc_cache *cache);
int thinc_cache_query_id(thinc_cache *cache, unsigned long int h);
int thinc_cache_query_data(thinc_cache *cache, register unsigned char *data, 
	register unsigned long int length, unsigned long int *hash);
unsigned long int thinc_cache_get_hash(thinc_cache *cache, 
	register unsigned char *data, register unsigned long int length);
unsigned long int thinc_cache_add(thinc_cache *cache, 
	register unsigned char *data, register unsigned long int length, 
	void *priv);
int thinc_cache_add_hash(thinc_cache *cache, unsigned long int hash,
	void *priv);
void thinc_cache_delete(thinc_cache *cache, unsigned long int hash);
void* thinc_cache_get(thinc_cache *cache, unsigned long int hash);
