/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * $Author: $
 *
 * resize functions
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 */
/* ***
 * fant resizing: from page 156 of Digital Image Warping by George Woldberg. 
 * Adapted to resize 3 color channels at the same time, and to find pixels in
 * columns
 * ***
 */

/* TODO: Can we do rows and columns in one pass? (no need for tmp) */

/* set tabstop 8 */

#include "thinc_resize.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>

#define T_RZ_TESTING

/* FIXME:
 * Doesn't work as it is. Only precomputes correctly for some scales
#define PRECOMP
 */

#ifdef T_RZ_TESTING
# include <assert.h>
#endif

/* clips value v to range [low,high] */
#define CLIP(v, low, high) ((v) < (low) ? (low) : ((v) > (high) ? (high):(v)))
#define CLIP_UP(v, high) ((v) > (high) ? (high):(v))

#define MAX_GRAY	0xff
#define MAX_2BYTES	0xffff

#define DOUBLE_BASE	128.0
#define DOUBLE_ROUND	0.5

#define FP_SHIFT	7U	
#define FP_ROUND	64
#define FP_ONE		(1<<FP_SHIFT)

#define fp_in(d) \
	( (unsigned int) ( ( (d)*DOUBLE_BASE) + DOUBLE_ROUND) )

#define fp_out(fp) \
	( ( (fp) + FP_ROUND ) >> FP_SHIFT)

#define fp_mult(a, b) \
	( ( (a) * (b) ) >> FP_SHIFT)

/** resize using fant's resampling algorithm using fixed point arithmetic ***/

/* TODO: Make yuv use fixed point functions */
static int thinc_fant_simple_row8(unsigned char *in, unsigned char *out,
	unsigned int ineln, unsigned int outlen);
/* TODO: Make yuv use fixed point functions */
static int thinc_fant_simple_col8(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride);

static int thinc_fant_simple_row8_fp(unsigned char *in, unsigned char *out,
	unsigned int ineln, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale);
static int thinc_fant_simple_col8_fp(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, unsigned int fpscale,
	unsigned int fpinvscale);
#if 0
static int thinc_fant_simple_row16(unsigned char *in, unsigned char *out,
	unsigned int ineln, unsigned int outlen);
static int thinc_fant_simple_col16(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride);
#else
static int thinc_fant_simple_row16_fp(unsigned char *in, unsigned char *out,
	unsigned int ineln, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale);
static int thinc_fant_simple_col16_fp(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, unsigned int fpscale,
	unsigned int fpinvscale);
#endif
static int thinc_resizeyuv_help(unsigned char *in, unsigned char *out,
	unsigned int inw, unsigned int inh, unsigned int outw, 
	unsigned int outh, unsigned char first);

static int thinc_fant_simple_row32_fp(unsigned char *in, unsigned char *out,
	unsigned int ineln, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale);

static int thinc_fant_simple_col32_fp(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, unsigned int fpscale,
	unsigned int fpinvscale);

/***** Other possible resizing functions **********/
#if 0
static int thinc_fant_simple_row32(unsigned char *in, unsigned char *out,
	unsigned int ineln, unsigned int outlen);
static int thinc_fant_simple_col32(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, double scale);
static int thinc_interpol_row32(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, double scale);
static int thinc_paul_row32(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, double scale);
#endif


#if 0
int thinc_resize8(unsigned char *in, unsigned char *out, unsigned int inw, 
	unsigned int inh, unsigned int outw, unsigned outh, 
	unsigned int instride, unsigned int outstride)
{
	const int bpp = 1;
	register unsigned int i;
	register unsigned char *inc, *outc;
	unsigned char *tmp;
	unsigned int tmplen, inlen, outlen, tmpstride;
	int free_tmp = T_NO;
	
	inlen = inw*inh*bpp;
	outlen = outw*outh*bpp;

	/* rows first */
	if(inw == outw) {
		tmp = in;
		tmplen = inlen;
		tmpstride = instride;
	} else {
		/* tmp is inh*outw */
		tmpstride = outw*bpp;
		tmplen = tmpstride*inh;
		tmp = malloc(tmplen);
		if(!tmp)
			return T_ERROR;
		free_tmp = T_YES;
		inc = in;
		outc = tmp;
		for(i=0; i<inh; ++i) {
			thinc_fant_simple_row8(inc, outc, inw, outw);
			inc += instride;
			outc += tmpstride;
		}
	}
	/* columns */
	if(inh == outh) {
		memcpy(out, tmp, tmplen);
	} else {
		inc = tmp;
		outc = out;
		for(i=0; i<outw; ++i) {
			thinc_fant_simple_col8(inc, outc, inh, outh, 
				tmpstride, outstride);
			++inc;	/* 8 bits */
			++outc;
		}
	}

	if(free_tmp)
		free(tmp);

	return T_OK;
}
#else
int thinc_resize8(unsigned char *in, unsigned char *out, unsigned int inw, 
	unsigned int inh, unsigned int outw, unsigned outh, 
	unsigned int instride, unsigned int outstride)
{
	const int bpp = 1;
	register unsigned int i;
	register unsigned char *inc, *outc;
	unsigned char *tmp;
	unsigned int tmplen, inlen, outlen, tmpstride;
	int free_tmp = T_NO;
	double scale, invscale;
	unsigned int fpscale, fpinvscale;
	
	inlen = inw*inh*bpp;
	outlen = outw*outh*bpp;

	/* rows first */
	if(inw == outw) {
		tmp = in;
		tmplen = inlen;
		tmpstride = instride;
	} else {
		/* tmp is inh*outw */
		tmpstride = outw*bpp;
		tmplen = tmpstride*inh;
		tmp = malloc(tmplen);
		if(!tmp)
			return T_ERROR;
		free_tmp = T_YES;
		inc = in;
		outc = tmp;
		scale = (double)outw/(double)inw;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);

		for(i=0; i<inh; ++i) {
			thinc_fant_simple_row8_fp(inc, outc, inw, outw, 
				fpscale, fpinvscale);
			inc += instride;
			outc += tmpstride;
		}
	}
	/* columns */
	if(inh == outh) {
		/* TODO: Why not make out be tmp right away??? */
		memcpy(out, tmp, tmplen);
	} else {
		inc = tmp;
		outc = out;
		scale = (double)outh/(double)inh;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);
	
		for(i=0; i<outw; ++i) {
			thinc_fant_simple_col8_fp(inc, outc, inh, outh, 
				tmpstride, outstride, fpscale, fpinvscale);
			++inc;	/* 8 bits */
			++outc;
		}
	}

	if(free_tmp)
		free(tmp);

	return T_OK;
}
#endif

#if 0
int thinc_resize16(unsigned char *in, unsigned char *out, unsigned int inw, 
	unsigned int inh, unsigned int outw, unsigned outh, 
	unsigned int instride, unsigned int outstride)
{
	const int bpp = 2;
	register unsigned int i;
	register unsigned char *inc, *outc;
	unsigned char *tmp;
	unsigned int tmplen, inlen, outlen, tmpstride;
	int free_tmp = T_NO;
	
	inlen = inw*inh*bpp;
	outlen = outw*outh*bpp;

	/* rows first */
	if(inw == outw) {
		tmp = in;
		tmplen = inlen;
		tmpstride = instride;
	} else {
		/* tmp is inh*outw */
		tmpstride = outw*bpp;
		tmplen = tmpstride*inh;
		tmp = malloc(tmplen);
		if(!tmp)
			return T_ERROR;
		free_tmp = T_YES;
		inc = in;
		outc = tmp;
		for(i=0; i<inh; ++i) {
			thinc_fant_simple_row16(inc, outc, inw, outw);
			inc += instride;
			outc += tmpstride;
		}
	}
	/* columns */
	if(inh == outh) {
		memcpy(out, tmp, tmplen);
	} else {
		inc = tmp;
		outc = out;
		for(i=0; i<outw; ++i) {
			thinc_fant_simple_col16(inc, outc, inh, outh, 
				tmpstride, outstride);
			++inc;	/* 8 bits */
			++outc;
		}
	}

	if(free_tmp)
		free(tmp);

	return T_OK;
}
#else
int thinc_resize16(unsigned char *in, unsigned char *out, unsigned int inw, 
	unsigned int inh, unsigned int outw, unsigned outh, 
	unsigned int instride, unsigned int outstride)
{
	const int bpp = 2;
	register unsigned int i;
	register unsigned char *inc, *outc;
	unsigned char *tmp;
	unsigned int tmplen, inlen, outlen, tmpstride;
	int free_tmp = T_NO;
	double scale, invscale;
	unsigned int fpscale, fpinvscale;
	
	inlen = inw*inh*bpp;
	outlen = outw*outh*bpp;

	/* rows first */
	if(inw == outw) {
		tmp = in;
		tmplen = inlen;
		tmpstride = instride;
	} else {
		/* tmp is inh*outw */
		tmpstride = outw*bpp;
		tmplen = tmpstride*inh;
		tmp = malloc(tmplen);
		if(!tmp)
			return T_ERROR;
		free_tmp = T_YES;
		inc = in;
		outc = tmp;
		scale = (double)outw/(double)inw;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);

		for(i=0; i<inh; ++i) {
			thinc_fant_simple_row16_fp(inc, outc, inw, outw, 
				fpscale, fpinvscale);
			inc += instride;
			outc += tmpstride;
		}
	}
	/* columns */
	if(inh == outh) {
		/* TODO: Why not make out be tmp right away??? */
		memcpy(out, tmp, tmplen);
	} else {
		inc = tmp;
		outc = out;
		scale = (double)outh/(double)inh;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);
	
		for(i=0; i<outw; ++i) {
			thinc_fant_simple_col16_fp(inc, outc, inh, outh, 
				tmpstride, outstride, fpscale, fpinvscale);
			inc += 2;	/* 32 bits */
			outc += 2;
		}
	}

	if(free_tmp)
		free(tmp);

	return T_OK;
}
#endif

int thinc_resize32(unsigned char *in, unsigned char *out, unsigned int inw, 
	unsigned int inh, unsigned int outw, unsigned outh, 
	unsigned int instride, unsigned int outstride)
{
	const int bpp = 4;
	register unsigned int i;
	register unsigned char *inc, *outc;
	unsigned char *tmp;
	unsigned int tmplen, inlen, outlen, tmpstride;
	int free_tmp = T_NO;
	double scale, invscale;
	unsigned int fpscale, fpinvscale;
	
	inlen = inw*inh*bpp;
	outlen = outw*outh*bpp;

	/* rows first */
	if(inw == outw) {
		tmp = in;
		tmplen = inlen;
		tmpstride = instride;
	} else {
		/* tmp is inh*outw */
		tmpstride = outw*bpp;
		tmplen = tmpstride*inh;
		tmp = malloc(tmplen);
		if(!tmp)
			return T_ERROR;
		free_tmp = T_YES;
		inc = in;
		outc = tmp;
		scale = (double)outw/(double)inw;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);

		for(i=0; i<inh; ++i) {
			thinc_fant_simple_row32_fp(inc, outc, inw, outw, 
				fpscale, fpinvscale);
			inc += instride;
			outc += tmpstride;
		}
	}
	/* columns */
	if(inh == outh) {
		/* TODO: Why not make out be tmp right away??? */
		memcpy(out, tmp, tmplen);
	} else {
		inc = tmp;
		outc = out;
		scale = (double)outh/(double)inh;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);
	
		for(i=0; i<outw; ++i) {
			thinc_fant_simple_col32_fp(inc, outc, inh, outh, 
				tmpstride, outstride, fpscale, fpinvscale);
			inc += 4;	/* 32 bits */
			outc += 4;
		}
	}

	if(free_tmp)
		free(tmp);

	return T_OK;
}

int thinc_resizeyuv(
	unsigned char *iny, unsigned char *inu, unsigned char *inv,
	unsigned char *outy, unsigned char *outu,  unsigned char *outv,
	unsigned int inw, unsigned int inh, unsigned int outw, unsigned outh)
{
	unsigned int yinw, yinh, youtw, youth;
	unsigned int uvinw, uvinh, uvoutw, uvouth;
	int err;

	yinw = inw, yinh = inh;
	uvinw = inw >> 1;
	uvinh = inh >> 1;
	youtw = outw, youth = outh;
	uvoutw = outw >> 1;
	uvouth = outh >> 1;

	/* one plane at a time */
	err = thinc_resizeyuv_help(iny, outy, yinw, yinh, youtw, youth,
		T_TRUE);
	if(err == T_ERROR)
		return T_ERROR;
	err = thinc_resizeyuv_help(inu, outu, uvinw, uvinh, uvoutw, uvouth, 
		T_FALSE);
	if(err == T_ERROR)
		return T_ERROR;
	err = thinc_resizeyuv_help(inv, outv, uvinw, uvinh, uvoutw, uvouth, 
		T_FALSE);
	if(err == T_ERROR)
		return T_ERROR;

	return T_OK;
}

/* the idea is to call this function sequentially for each of the three
 * channels. The first time (y) pass T_TRUE so that temporary memory
 * is allocated. Since the y channel is the largest one it's ok to use
 * this memory to resize the u & v channels */
static int thinc_resizeyuv_help(unsigned char *in, unsigned char *out,
	unsigned int inw, unsigned int inh, unsigned int outw, 
	unsigned int outh, unsigned char first)
{
	static unsigned char *tmp = NULL;
	register unsigned char *inc, *outc;
	register unsigned int i;

	if(inw == inh) {
		tmp = in;
	} else {
		if(first) {
			if(tmp)
				free(tmp);
			tmp = malloc(outw*inh);
			if(!tmp)
				return T_ERROR;
		}
		inc = in;
		outc = tmp;
		for(i=0; i<inh; ++i) {
			thinc_fant_simple_row8(inc, outc, inw, outw);
			inc += inw;
			outc += outw;
		}
	}

	if(outw == outh) {
		memcpy(out, tmp, outw*outh);
	} else {
		inc = tmp;
		outc = out;
		for(i=0; i<outw; ++i) {
			thinc_fant_simple_col8(inc, outc, inh, outh,
				outw, outw);
			++inc;
			++outc;
		}
	}

	return T_OK;
}

static int thinc_fant_simple_row8(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen)
{
	unsigned int u,x;
	double 	scale, invscale, inleft,
		outleft; /* # input pixels that map onto 1 output pixel */ 
	double intensity;
	double acc;

	scale = (double)outlen/(double)inlen;
	invscale = 1.0/scale;
	outleft = invscale;
	inleft = 1.0;	/* entire input pixel is available */
	acc = 0.0;

	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index */
			--u;
		}
		/* use linear interpolation for reconstruction */
		intensity = (inleft * in[u]) + ((1-inleft)*in[u+1]);

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			acc += (intensity *inleft); /*weigthed contribution*/
			outleft -= inleft; /*inleft portion has been filled*/
			inleft = 1.0;
			++u;
		} else {
		/* inleft >= outleft: input pixel not entirely consumed
		 * before output pixel */
			acc += (intensity * outleft);
			out[x] = acc * scale;
			acc = 0.0;
			inleft -= outleft;
			outleft = invscale;
			++x;
		}
	}

	return T_OK;
}

/* do it column by column (doesn't know how to go to next column), and 
 * inlen and outlen should be the height in pixels, not bytes. stride 
 * should be in bytes */
static int thinc_fant_simple_col8(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride)
{
	unsigned int u,x, ui, xi;
	double 	scale, invscale, inleft,
		outleft; /* # input pixels that map onto 1 output pixel */ 
	double intensity;
	double acc;

	scale = (double)outlen/(double)inlen;
	invscale = 1.0/scale;
	outleft = invscale;
	inleft = 1.0;	/* entire input pixel is available */
	acc = 0.0;

	ui = 0;
	xi = 0;
	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index. just go back one line and look the
			 * other way (when magnifying it does produce some
			 * repetition artifacts) */
			--u;
			ui -= instride;
		}
		/* use linear interpolation for reconstruction */
		intensity = (inleft * in[ui]) + ((1-inleft)*in[ui+instride]);

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			acc += (intensity *inleft); /*weigthed contribution*/
			outleft -= inleft; /*inleft portion has been filled*/
			inleft = 1.0;
			++u;
			ui += instride;
		} else {
		/* inleft >= outleft: input pixel not entirely consumed
		 * before output pixel */
			acc += (intensity * outleft);
			out[xi] = acc * scale;
			acc = 0.0;
			inleft -= outleft;
			outleft = invscale;
			++x;
			xi += outstride;
		}
	}

	return T_OK;
}

static int thinc_fant_simple_row8_fp(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int intensity;
	unsigned int acc;

	register unsigned char *inp, *outp;

	outleft = fpinvscale;
	inleft = FP_ONE;
	acc = 0;

	inp = in;
	outp = out;

	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index */
			--u;
			--inp;
		}
		/* use linear interpolation for reconstruction */
		intensity = inleft * inp[0] + (FP_ONE-inleft) * inp[1];

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			acc += fp_mult(intensity, inleft); 

			outleft -= inleft; /*inleft portion has been filled*/
			inleft = FP_ONE;
			++u;
			++inp;
		} else {
			/* inleft >= outleft: input pixel not entirely 
			 * consumed before output pixel */
			acc += fp_mult(intensity, outleft); 

			outp[0] = CLIP_UP(
				fp_out(fp_mult(acc, fpscale)), MAX_GRAY);
			acc = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp++;
		}
	}

	return T_OK;
}

static int thinc_fant_simple_col8_fp(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int intensity;
	unsigned int acc;

	unsigned char *inp, *ninp, *outp;

	outleft = fpinvscale;
	inleft = FP_ONE;
	acc = 0;

	inp = in;
	ninp = in+instride;
	outp = out;
	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index. just go back one line and look the
			 * other way (when magnifying it does produce some
			 * repetition artifacts) */
			--u;
			ninp = inp;
			inp -= instride;
		}
		/* use linear interpolation for reconstruction */
		intensity = inleft * inp[0] + 
			(FP_ONE-inleft) * ninp[0];

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			acc += fp_mult(intensity, inleft); 
			outleft -= inleft; /*inleft portion has been filled*/

			inleft = FP_ONE;
			++u;
			inp = ninp;
			ninp += instride;
		} else {
			/* inleft >= outleft: input pixel not entirely consumed
			 * before output pixel */
			acc += fp_mult(intensity, outleft); 

			outp[0] = CLIP_UP(
					fp_out(fp_mult(acc, fpscale)), 
					MAX_GRAY);

			acc = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp += outstride;
		}
	}

	return T_OK;
}

#if 0
/* we treat the whole pixel as a single entity (as opposed to 24bpp where we
 * separate each channel. separing the three channels would be just too
 * expensive. hopefully this still gives us good results */
static int thinc_fant_simple_row16(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen)
{
	unsigned int u,x;
	double 	scale, invscale, inleft,
		outleft; /* # input pixels that map onto 1 output pixel */ 
#if 0
	double intensity;
	double acc;
	/* TODO: Assuming short is 16 bits. isn't there a u16 type or
	 * something like that? */
	unsigned short *in_16, *out_16;
#else
	double intensity1, intensity2;
	double acc1, acc2;
#endif

	scale = (double)outlen/(double)inlen;
	invscale = 1.0/scale;
	outleft = invscale;
	inleft = 1.0;	/* entire input pixel is available */
#if 0
	acc = 0.0;
	in_16 = (unsigned short*) in;
	out_16 = (unsigned short*) out;
#else
	acc1 = acc2 = 0.0;
#endif

	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index */
			--u;
		}
		/* use linear interpolation for reconstruction */
#if 0
		intensity = (inleft * in_16[u]) + ((1-inleft)*in_16[u+1]);
#else
		intensity1 = (inleft * in[u]) + ((1-inleft)*in[u+2]);
		intensity2 = (inleft * in[u+1]) + ((1-inleft)*in[u+3]);

#endif

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
#if 0
			acc += (intensity *inleft); /*weigthed contribution*/
#else
			acc1 += (intensity1 *inleft); /*weigthed contribution*/
			acc2 += (intensity2 *inleft); /*weigthed contribution*/
#endif
			outleft -= inleft; /*inleft portion has been filled*/
			inleft = 1.0;
			++u;
		} else {
			/* inleft >= outleft: input pixel not entirely consumed
			 * before output pixel */
#if 0
			acc += (intensity * outleft);
			out_16[x] = acc * scale;
			acc = 0.0;
#else
			acc1 += (intensity1 * outleft);
			acc2 += (intensity2 * outleft);
			out[x] = acc1 * scale;
			out[x+1] = acc2 * scale;
			acc1 = acc2 = 0.0;
#endif
			inleft -= outleft;
			outleft = invscale;
			++x;
		}
	}

	return T_OK;
}

/* do it column by column (doesn't know how to go to next column), and 
 * inlen and outlen should be the height in pixels, not bytes. stride 
 * should be in bytes */
static int thinc_fant_simple_col16(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride)
{
	unsigned int u,x, ui, xi;
	double 	scale, invscale, inleft,
		outleft; /* # input pixels that map onto 1 output pixel */ 
#if 0
	double intensity;
	double acc;
	/* TODO: Assuming short is 16 bits. isn't there a u16 type or
	 * something like that? */
	unsigned short *in_16, *out_16;
#else
	double intensity1, intensity2;
	double acc1, acc2;
#endif

	scale = (double)outlen/(double)inlen;
	invscale = 1.0/scale;
	outleft = invscale;
	inleft = 1.0;	/* entire input pixel is available */
#if 0
	acc = 0.0;
	in_16 = (unsigned short*) in;
	out_16 = (unsigned short*) out;
#else
	acc1 = acc2 = 0.0;
#endif

	ui = 0;
	xi = 0;

	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index. just go back one line and look the
			 * other way (when magnifying it does produce some
			 * repetition artifacts) */
			--u;
			ui -= instride;
		}
		/* use linear interpolation for reconstruction */
#if 0
		intensity = (inleft * in_16[ui]) + 
			((1-inleft)*in_16[ui+instride]);
#else
		intensity1 = (inleft * in[ui]) + 
			((1-inleft)*in[ui+instride]);
		intensity2 = (inleft * in[ui+1]) + 
			((1-inleft)*in[ui+instride+1]);
#endif

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
#if 0
			acc += (intensity *inleft); /*weigthed contribution*/
#else
			acc1 += (intensity1 *inleft); /*weigthed contribution*/
			acc2 += (intensity2 *inleft); /*weigthed contribution*/
#endif
			outleft -= inleft; /*inleft portion has been filled*/
			inleft = 1.0;
			++u;
			ui += instride;
		} else {
		/* inleft >= outleft: input pixel not entirely consumed
		 * before output pixel */
#if 0
			acc += (intensity * outleft);
			out_16[xi] = acc * scale;
			acc = 0.0;
#else
			acc1 += (intensity1 * outleft);
			acc2 += (intensity2 * outleft);
			out[xi] = acc1 * scale;
			out[xi+1] = acc2 * scale;
			acc1 = acc2 = 0.0;
#endif
			inleft -= outleft;
			outleft = invscale;
			++x;
			xi += outstride;
		}
	}

	return T_OK;
}
#else
static int thinc_fant_simple_row16_fp(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int intensity1, intensity2;
	unsigned int acc1, acc2;

#if 0
	register unsigned char *inp, *outp;
#else
	register unsigned short *inp, *outp;
#endif

	outleft = fpinvscale;
	inleft = FP_ONE;
	acc1 = acc2 = 0;

#if 0
	inp = in;
	outp = out;
#else
	inp = (unsigned short*) in;
	outp = (unsigned short*) out;
#endif

	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index */
			--u;
#if 0
			inp -= 2;
#else
			--inp;
#endif
		}
#if 0
		/* use linear interpolation for reconstruction */
		intensity1 = inleft * inp[0] + 
			(FP_ONE-inleft) * inp[2];
		intensity2 = inleft * inp[1] + 
			(FP_ONE-inleft) * inp[3];
#else
		intensity1 = inleft * inp[0] + (FP_ONE-inleft)*inp[1];
#endif

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
#if 0
			acc1 += fp_mult(intensity1, inleft); 
			acc2 += fp_mult(intensity2, inleft);
#else
			acc1 += fp_mult(intensity1, inleft); 
#endif

			outleft -= inleft; /*inleft portion has been filled*/
			inleft = FP_ONE;
			++u;
#if 0
			inp += 2;
#else
			inp++;
#endif
		} else {
			/* inleft >= outleft: input pixel not entirely 
			 * consumed before output pixel */
#if 0
			acc1 += fp_mult(intensity1, outleft); 
			acc2 += fp_mult(intensity2, outleft);
			outp[0] = CLIP_UP(
				fp_out(fp_mult(acc1, fpscale)), MAX_GRAY);
			outp[1] = CLIP_UP(
				fp_out(fp_mult(acc2, fpscale)), MAX_GRAY);
#else
			acc1 += fp_mult(intensity1, outleft); 
			outp[0] = CLIP_UP(
				fp_out(fp_mult(acc1, fpscale)), MAX_2BYTES);
#endif

			acc1 = acc2 = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
#if 0
			outp+= 2;
#else
			outp++;
#endif
		}
	}

	return T_OK;
}

static int thinc_fant_simple_col16_fp(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int intensity1, intensity2;
	unsigned int acc1, acc2;

	unsigned char *inp, *ninp, *outp;

	outleft = fpinvscale;
	inleft = FP_ONE;
	acc1 = acc2 = 0;

	inp = in;
	ninp = in+instride;
	outp = out;
	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index. just go back one line and look the
			 * other way (when magnifying it does produce some
			 * repetition artifacts) */
			--u;
			ninp = inp;
			inp -= instride;
		}
		/* use linear interpolation for reconstruction */
		intensity1 = inleft * inp[0] + 
			(FP_ONE-inleft) * ninp[0];
		intensity2 = inleft * inp[1] + 
			(FP_ONE-inleft) * ninp[1];

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			acc1 += fp_mult(intensity1, inleft); 
			acc2 += fp_mult(intensity2, inleft);
			outleft -= inleft; /*inleft portion has been filled*/

			inleft = FP_ONE;
			++u;
			inp = ninp;
			ninp += instride;
		} else {
			/* inleft >= outleft: input pixel not entirely consumed
			 * before output pixel */
			acc1 += fp_mult(intensity1, outleft); 
			acc2 += fp_mult(intensity2, outleft);

			outp[0] = CLIP_UP(
					fp_out(fp_mult(acc1, fpscale)), 
					MAX_GRAY);
			outp[1] = CLIP_UP(
					fp_out(fp_mult(acc2, fpscale)), 
					MAX_GRAY);

			acc1 = acc2 = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp += outstride;
		}
	}

	return T_OK;
}
#endif

static int thinc_fant_simple_row32_fp(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int rintensity, gintensity, bintensity;
	unsigned int racc, gacc, bacc;

	register unsigned char *inp, *outp;

	outleft = fpinvscale;
	inleft = FP_ONE;
	racc = gacc = bacc = 0;

	inp = in;
	outp = out;

	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index */
			--u;
			inp -= 4;
		}
		/* use linear interpolation for reconstruction */
		rintensity = inleft * inp[0] + 
			(FP_ONE-inleft) * inp[4];
		gintensity = inleft * inp[1] + 
			(FP_ONE-inleft) * inp[5];
		bintensity = inleft*inp[2] + 
			(FP_ONE-inleft) * inp[6];

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			racc += fp_mult(rintensity, inleft); 
			gacc += fp_mult(gintensity, inleft);
			bacc += fp_mult(bintensity, inleft);

			outleft -= inleft; /*inleft portion has been filled*/
			inleft = FP_ONE;
			++u;
			inp += 4;
		} else {
			/* inleft >= outleft: input pixel not entirely 
			 * consumed before output pixel */
			racc += fp_mult(rintensity, outleft); 
			gacc += fp_mult(gintensity, outleft);
			bacc += fp_mult(bintensity, outleft);

			outp[0] = CLIP_UP(
				fp_out(fp_mult(racc, fpscale)), MAX_GRAY);
			outp[1] = CLIP_UP(
				fp_out(fp_mult(gacc, fpscale)), MAX_GRAY);
			outp[2] = CLIP_UP(
				fp_out(fp_mult(bacc, fpscale)), MAX_GRAY);

			racc = gacc = bacc = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp+= 4;
		}
	}

	return T_OK;
}

static int thinc_fant_simple_col32_fp(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int rintensity, gintensity, bintensity;
	unsigned int racc, gacc, bacc;

	unsigned char *inp, *ninp, *outp;

	outleft = fpinvscale;
	inleft = FP_ONE;
	racc = gacc = bacc = 0;

	inp = in;
	ninp = in+instride;
	outp = out;
	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index. just go back one line and look the
			 * other way (when magnifying it does produce some
			 * repetition artifacts) */
			--u;
			ninp = inp;
			inp -= instride;
		}
		/* use linear interpolation for reconstruction */
		rintensity = inleft * inp[0] + 
			(FP_ONE-inleft) * ninp[0];
		gintensity = inleft * inp[1] + 
			(FP_ONE-inleft) * ninp[1];
		bintensity = inleft*inp[2] + 
			(FP_ONE-inleft) * ninp[2];

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			racc += fp_mult(rintensity, inleft); 
			gacc += fp_mult(gintensity, inleft);
			bacc += fp_mult(bintensity, inleft);
			outleft -= inleft; /*inleft portion has been filled*/

			inleft = FP_ONE;
			++u;
			inp = ninp;
			ninp += instride;
		} else {
			/* inleft >= outleft: input pixel not entirely consumed
			 * before output pixel */
			racc += fp_mult(rintensity, outleft); 
			gacc += fp_mult(gintensity, outleft);
			bacc += fp_mult(bintensity, outleft);

			outp[0] = CLIP_UP(
					fp_out(fp_mult(racc, fpscale)), 
					MAX_GRAY);
			outp[1] = CLIP_UP(
					fp_out(fp_mult(gacc, fpscale)), 
					MAX_GRAY);
			outp[2] = CLIP_UP(
					fp_out(fp_mult(bacc, fpscale)), 
					MAX_GRAY);

			racc = gacc = bacc = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp += outstride;
		}
	}

	return T_OK;
}

/************ Other possible resizing functions **************/
#if 0
static int thinc_fant_simple_row32(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen)
{
	unsigned int u,x, ui, xi;
	double 	scale, invscale, inleft,
		outleft; /* # input pixels that map onto 1 output pixel */ 
	double rintensity, gintensity, bintensity;
	double racc, gacc, bacc;

	scale = (double)outlen/(double)inlen;
	invscale = 1.0/scale;
	outleft = invscale;
	inleft = 1.0;	/* entire input pixel is available */
	racc = gacc = bacc = 0.0;

	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index */
			--u;
		}
		/* use linear interpolation for reconstruction */
		ui = u<<2;
		rintensity = (inleft * in[ui]) + ((1-inleft)*in[ui+4]);
		gintensity = (inleft * in[ui+1]) + ((1-inleft)*in[ui+5]);
		bintensity = (inleft * in[ui+2]) + ((1-inleft)*in[ui+6]);

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			racc += (rintensity *inleft); /*weigthed contribution*/
			gacc += (gintensity *inleft);
			bacc += (bintensity *inleft);
			outleft -= inleft; /*inleft portion has been filled*/
			inleft = 1.0;
			++u;
		} else {
			/* inleft >= outleft: input pixel not entirely 
			 * consumed before output pixel */
			racc += (rintensity * outleft);
			gacc += (gintensity * outleft);
			bacc += (bintensity * outleft);
			xi = x<<2;
			out[xi] = racc * scale;
			out[xi+1] = gacc * scale;
			out[xi+2] = bacc * scale;
			racc = gacc = bacc = 0.0;
			inleft -= outleft;
			outleft = invscale;
			++x;
		}
	}

	return T_OK;
}

static int thinc_fant_simple_col32(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, double scale)
{
	unsigned int u,x, ui, xi;
	double 	invscale, inleft,
		outleft; /* # input pixels that map onto 1 output pixel */ 
	double rintensity, gintensity, bintensity;
	double racc, gacc, bacc;

	invscale = 1.0/scale;
	outleft = invscale;
	inleft = 1.0;	/* entire input pixel is available */
	racc = gacc = bacc = 0.0;

	ui = 0;
	xi = 0;
	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index. just go back one line and look the
			 * other way (when magnifying it does produce some
			 * repetition artifacts) */
			--u;
			ui -= instride;
		}
		/* use linear interpolation for reconstruction */
		rintensity = (inleft * in[ui]) + ((1-inleft)*in[ui+instride]);
		gintensity = (inleft * in[ui+1]) + 
			((1-inleft)*in[ui+instride+1]);
		bintensity = (inleft * in[ui+2]) + 
			((1-inleft)*in[ui+instride+2]);

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			racc += (rintensity *inleft); /*weigthed contribution*/
			gacc += (gintensity *inleft);
			bacc += (bintensity *inleft);
			outleft -= inleft; /*inleft portion has been filled*/
			inleft = 1.0;
			++u;
			ui += instride;
		} else {
		/* inleft >= outleft: input pixel not entirely consumed
		 * before output pixel */
			racc += (rintensity * outleft);
			gacc += (gintensity * outleft);
			bacc += (bintensity * outleft);
			out[xi] = racc * scale;
			out[xi+1] = gacc * scale;
			out[xi+2] = bacc * scale;
			racc = gacc = bacc = 0.0;
			inleft -= outleft;
			outleft = invscale;
			++x;
			xi += outstride;
		}
	}

	return T_OK;
}

static int thinc_interpol_row32(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, double scale)
{
	/* x is center of kernel. u is position in input */
	double x, u, fval, weight;
	/* per channel data */
	double rsum, gsum, bsum;
	double rval, gval, bval;
	/* left and right indices of kernel */
	int l, r;
	int i;
	double kwidth;			/* width of the kernel */
	/* if we're minifying we need to make the kernel broader, and
	 * therefore shallower */
	double kscale;
	register unsigned char *inp, *outp;

	kwidth = 1.0;
	if(scale < 1.0) {
		/* minifying, make kernel broader and shallower */
		kwidth /= scale;
		kscale = scale;
	} else
		kscale = 1.0;

	outp = out;

	/* center the kernel in all x's and find output value */
	for(x=0; x<outlen; ++x) {
		/* u is center of kernel in input */
		u = x/scale;

		/* find kernel's endpoints */
		l = (int)ceil(u-kwidth);
		r = (int)floor(u+kwidth);
		l = CLIP(l, 0, inlen-1);
		r = CLIP(r, 0, inlen-1);

		rsum = 0;
		gsum = 0;
		bsum = 0;

		inp = &in[l<<2];

		for(i=l; i<=r; ++i) {
			rval = *inp++;
			gval = *inp++;
			bval = *inp++;
			++inp;

			/* in case of minifying, we need to put the distance 
			 * back to the range of values understood by the 
			 * filter (remember we had to make the kernel 
			 * broader), so we multiply it by 
			 * kscale = 1/broadening factor */
			fval = (u-i)*kscale;
			weight = 1.0 - fabs(fval);

			rsum += weight*rval;
			gsum += weight*gval;
			bsum += weight*bval;
		}

		/* in case of minifying, the kernel's amplitude must be lower
		 * (inversely proportional to minifying factor) */
		*outp++ = (unsigned char) CLIP((int)(rsum * kscale), 0, 
			MAX_GRAY);
		*outp++ = (unsigned char) CLIP((int)(gsum * kscale), 0, 
			MAX_GRAY);
		*outp++ = (unsigned char) CLIP((int)(bsum * kscale), 0, 
			MAX_GRAY);
		*outp++ = 0xFF;
	}

	return T_OK;
}

static int thinc_paul_row32(unsigned char *in, unsigned char *out,
	unsigned int inlen, unsigned int outlen, double scale)
{
	long i, j, n, m, d, adj;
	long acc_r, acc_g, acc_b;
	long acc_2r, acc_2g, acc_2b;
	long in_r, in_g, in_b;

	j=0;
	n=inlen;
	m=outlen;
	acc_r=acc_g=acc_b=0;
	acc_2r=acc_2g=acc_2b=0;

#define ADJ_DIGITS 16
	adj=(long)((((double)m)/((double)(n*n)))*((double)(1L<<ADJ_DIGITS)));

	for (i=0; i<n && j<m;) {
		d=abs(i*m-j*n);
		if (d<=n) {
			in_r=in[i*4];
			in_g=in[i*4+1];
			in_b=in[i*4+2];
			acc_r+=in_r*((n-d));
			acc_g+=in_g*((n-d));
			acc_b+=in_b*((n-d));
			acc_2r+=in_r*(d);
			acc_2g+=in_g*(d);
			acc_2b+=in_b*(d);
			i++;
		}
		else {
			acc_r*=adj;
			acc_r>>=ADJ_DIGITS;
			acc_g*=adj;
			acc_g>>=ADJ_DIGITS;
			acc_b*=adj;
			acc_b>>=ADJ_DIGITS;

			out[j*4]=(acc_r>255)?255:acc_r;
			out[j*4+1]=(acc_g>255)?255:acc_g;
			out[j*4+2]=(acc_b>255)?255:acc_b;
			acc_r=acc_2r;
			acc_g=acc_2g;
			acc_b=acc_2b;
			j++;
			acc_2r=acc_2g=acc_2b=0;
		}
	}
	return T_OK;
}
#endif
