/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/******************************Module*Header*******************************\
*
*                           *********************
*                           * THINC header file *
*                           *********************
*
* Module Name: command.h
*
* contains the defination shared between the kernel mode driver and user mode
* process. Please be very careful in changing it. Otherwise, it will crash
* the system easily. 
*
***************************************************************************/

#ifndef __COMMON__
#define __COMMON__

#define GESC_SERVER_START		0x21000
#define GESC_START_COPY_CMD		0x22000
#define GESC_FINISH_COPY_CMD	0x23000
#define GESC_START_COPY_RAW		0x24000
#define GESC_FINISH_COPY_RAW	0x25000
#define GESC_RECREATE_FB		0x26000

#define MAX_OFF_SURF	2000
#define MIN_OFF_SURF	2
#define FRAMEBUFF_ID	1
#define NULL_SURF_ID	0

#if DBG
#define TIME_STAMP
#endif 

typedef struct _mirror_cursor {
	int cursorSize; /*	If positive, has the size of the cursor in 
  						bytes (the source and the mask each are
						of size cursorSize)
						If -1, no cursor handling will be
						done by the client and the rest of the
						information in this message should be 
						disregarded */
	int cursorWidth;/* cursor size in pixels */
	int cursorHeight;
	int xHot;
	int yHot;
} mirror_cursor;

typedef struct MapToUserOne_s {
	PVOID	userVirtualAddress;
	INT		flags;
} MapToUserOne;

typedef struct MapToUser_s {
	INT	len;
	MapToUserOne MapToUserList[4];		// the actual length of the list is indicated by len
										// I think len will never be greater 1000
}  MapToUserMode, *PMapToUserMode;

#define MM_FG_ON_SURF	0x0001
#define MM_FG_OFF_SURF	0x0002
#define MM_FG_CMD_BLK	0x0004
#define MM_FG_CMD_QUE	0x0008
#define MM_FG_SIGNAL	0x1000
#define MM_FG_OTHER		0x2000

typedef struct ThincSignals_s{
	// Set by user mode server
	BOOL	bReqBlockDrawing;
	// Set by driver
	BOOL	bStatusBlockBeforeCopy;
	BOOL	bStatusDrawing;
	BOOL	bStatusWaitingForDrawingFinishes;
	BOOL	bReqNewCommandSignal;
} ThincSignals, *PThincSignals;

typedef struct ParametersIn_s{
	INT		recheckInterval;
} ParaIn, *PParaIn;

typedef PVOID MemToUnmap;		// a handle to the mapped block;

// Flags for MirrSurf
#define SURF_FG_ONSCREEN	0x01

#define T_CREATE_OFFSURF 50 /* thinc_createPixmap */
#define T_DESTROY_OFFSURF 51 /* thinc_destroyOffscreenSurf */

typedef struct thinc_createOffscreenSurf_s {
	int width;	/* fb dimensions */
	int height;
	int depth;		/* color depth of fb */
	int bpp;		/* how many bytes are used for each pixel */	
	int id;		/* id of the surface */
	int stride;
} thinc_createOffscreenSurf;

#define T_CREATE_OFFSURF_SIZE	sizeof(thinc_createOffscreenSurf)

typedef struct thinc_destroyOffscreenSurf_s {
	int	id;
} thinc_destroyOffscreenSurf;

#define T_DESTROY_OFFSURF_SIZE	sizeof(thinc_destroyOffscreenSurf)

#define T_TIME_STAMP				52  /* thinc_timeStamp */
#define TIME_STAMP_ORIG				0x01
#define TIME_STAMP_THINC			0x02
#define TIME_STAMP_BEFORE_WAIT		0x04
#define TIME_STAMP_AFTER_WAIT		0x08
#define TIME_STAMP_DRIVER_FUNCTION	0x10
#define TIME_STAMP_AFTER_DRAWING	0x20

typedef struct thinc_timeStamp_s{
	char time[16];	
	int stampType;
	char record[32];
} thinc_timeStamp;
#define T_TIME_STAMP_SIZE sizeof(thinc_timeStamp);
#endif /* __COMMON__ */

