/******************************Module*Header*******************************\
*
*                           *******************
*                           * GDI SAMPLE CODE *
*                           *******************
*
* Module Name: driver.h
*
* contains prototypes for the frame buffer driver.
*
* Copyright (c) 1992-1998 Microsoft Corporation
\**************************************************************************/

#ifndef __DRIVER__
#define __DRIVER__

//#define DBG 1
#define TIME_STAMP
#include "stddef.h"

#include <stdarg.h>
//#include <string.h>

#include "windef.h"
#include "wingdi.h"
#include "winddi.h"
#include "devioctl.h"
#include "ntddvdeo.h"
#include "debug.h"
#include "utils.h"
#include "mem.h"
#include "common.h"

#define thincAlloc(size, tag) \
	EngAllocMem(FL_NONPAGED_MEMORY, size, MKTAG(tag))

#define thincAllocZero(size, tag) \
	EngAllocMem(FL_NONPAGED_MEMORY | FL_ZERO_MEMORY , size, MKTAG(tag))

#define thincFree(x) EngFreeMem(x)

typedef struct ThincCmdBlock_s ThincCmdBlock, *PThincCmdBlock;
typedef struct ThincDrvNode_s ThincDrvNode, *PThincDrvNode, ThincDrvList, *PThincDrvList;
typedef struct ThincCmdPtrQueue_s ThincCmdPtrQueue, *PThincCmdPtrQueue;
typedef struct MirrSurf_s MirrSurf, *PMirrSurf;

/* refer to \mirror\disp\driver.h@line: 27 */
/* refer to \mirror\disp\driver.h@line: 28 */
/* refer to \mirror\disp\driver.h@line: 29 */                    // Handle to \Device\Screen
/* refer to \mirror\disp\driver.h@line: 30 */                    // Engine's handle to PDEV
/* refer to \mirror\disp\driver.h@line: 31 */                   // Engine's handle to surface
/* refer to \mirror\disp\driver.h@line: 32 */               // Handle to the default palette for device.
/* refer to \mirror\disp\driver.h@line: 33 */                   // This is pointer to base screen address
/* refer to \mirror\disp\driver.h@line: 34 */                   // Visible screen width
/* refer to \mirror\disp\driver.h@line: 35 */                   // Visible screen height
/* refer to \mirror\disp\driver.h@line: 36 */                     // Where this display is anchored in
/* refer to \mirror\disp\driver.h@line: 37 */
/* refer to \mirror\disp\driver.h@line: 38 */                     // Mode the mini-port driver is in.
/* refer to \mirror\disp\driver.h@line: 39 */               // Distance from one scan to the next.
/* refer to \mirror\disp\driver.h@line: 40 */                // size of video memory, including
/* refer to \mirror\disp\driver.h@line: 41 */
/* refer to \mirror\disp\driver.h@line: 42 */             // linked list of DCI offscreen surfaces.
/* refer to \mirror\disp\driver.h@line: 43 */                      // For bitfields device, Red Mask
/* refer to \mirror\disp\driver.h@line: 44 */                    // For bitfields device, Green Mask
/* refer to \mirror\disp\driver.h@line: 45 */                     // For bitfields device, Blue Mask
/* refer to \mirror\disp\driver.h@line: 46 */              // number of bits the 8-8-8 palette must
/* refer to \mirror\disp\driver.h@line: 47 */
/* refer to \mirror\disp\driver.h@line: 48 */
/* refer to \mirror\disp\driver.h@line: 49 */                 // # of bits per pel 8,16,24,32 are only supported.
/* refer to \mirror\disp\driver.h@line: 50 */                 // adjustment for pointer hot spot
/* refer to \mirror\disp\driver.h@line: 51 */ // HW pointer abilities
/* refer to \mirror\disp\driver.h@line: 52 */ // hardware pointer attributes
/* refer to \mirror\disp\driver.h@line: 53 */        // Size of buffer allocated
/* refer to \mirror\disp\driver.h@line: 54 */            // Are we currently using the hw cursor
/* refer to \mirror\disp\driver.h@line: 55 */                 // If this is pal managed, this is the pal
/* refer to \mirror\disp\driver.h@line: 56 */                // Does the miniport support DCI?

/* refer to \mirror\disp\driver.h@line: 58 */                // ptr to MirrSurf bits for screen surface

	//THINC
	HANDLE			hMemDll;			// handle to the mem.dll image

	PFCreateMem		pfCreateMem;
	PFDeAllocMem	pfDeAllocMem;
	PFMapMem		pfMapMem;
	PFUnMapMem		pfUnMapMem;

	PFOpenEvent		pfOpenEvent;
	PFCloseEvent	pfCloseEvent;
	PFWaitOnEvent	pfWaitOnEvent;
	PFSetEvent		pfSetEvent;
	PFResetEvent	pfResetEvent;
	PFQuerySystemTime pfQuerySystemTime;

	PFAtomicCompareExchange pfAtomicCompareExchange;

	PVOID			pEventNewCmd;		// Set when a new command is created; 
	PVOID			pEventBlockOnDraw;	// Wait before any drawing;
	PVOID			pEventStartBlocking;// Set before waiting on pEventBlockOnDraw
	PVOID			pEventRestart;	
	PVOID			pEventBlockOnCopy;	// Wait before copy; set by user mode server
	PVOID			pEventStartBlockingCopy;// Set before waiting on pEventBlockOnCopy

	INT				iCallStackDepth;
	//PThincDrvList	pListToUnmap;
	//PThincDrvList	pListToMap;
	//PThincDrvList	pListZombieCmdBlock;
	//PThincDrvList	pListZombieSurf;
	//PThincDrvList	pListAcitveSurf;

	BOOL			bServerOn;
	//BOOL			bCopyingCmds;
	//BOOL			bCopyingRawData;
	ThincSignals	thincSignals;

	PMirrSurf		surfList[MAX_OFF_SURF];
	INT				lastOffScreenSurfId;

	PThincCmdPtrQueue	pCmdPtrQueue;			
	PThincCmdBlock		pCmdBlock;
    
	// mouse cursor
	BOOL	drvManagedCursor;
	INT		cursorWidth;
	INT		cursorHeight;
	INT		cursorSize;
	INT		xHot;
	INT		yHot;
	VOID*	cursorMask;					// The size of cursor mask is 2 * cxCursor * cyCursor in bytes
/* refer to \mirror\disp\driver.h@line: 59 */


// Flags for MirrSurf
#define SURF_FG_ONSCREEN	0x01
#define SURF_FG_ZOMBIE		0x02

typedef struct ThincCmdBlock_s ThincCmdBlock, *PThincCmdBlock;

typedef struct MirrSurf_s {
	// for driver only
	PVOID	userVirtualAddress;
	HANDLE	hShareMem;
	ULONG	flags;                                   
/* refer to \3dlabs\perm3\disp\inc\glddtk.h@line: 371 */
    INT		id;
	//PThincDrvNode	pListNode;	// pointing to the node by which the surface is linked 
	//							// to a certain list

	
	// share with MirrSurfUser, should be consistent in size.
	LONG	cx;               
    LONG	cy;
    LONG	lDelta;
/* refer to \mirror\disp\driver.h@line: 49 */
	PVOID	kernelVirtualAddress;
	ULONG	size;

	CHAR	pscan[0];
    // followed by size of bytes
/* refer to \mirror\disp\driver.h@line: 69 */

/* refer to \mirror\disp\driver.h@line: 71 */
/* refer to \mirror\disp\driver.h@line: 72 */
/* refer to \mirror\disp\driver.h@line: 73 */
/* refer to \mirror\disp\driver.h@line: 74 */
/* refer to \mirror\disp\driver.h@line: 75 */
/* refer to \mirror\disp\driver.h@line: 76 */
/* refer to \mirror\disp\driver.h@line: 77 */
/* refer to \mirror\disp\driver.h@line: 78 */
BOOL UnMapAndFreeMem(DRIVEROBJ *);

void BlockOnRequest(SURFOBJ* surfobj1, SURFOBJ* surfobj2);


INT		RawUpdate(SURFOBJ* psoDst, RECTL* rct);
INT		FakeRawUpdate(PMirrSurf	psurfDst);
INT		SolidFill(PMirrSurf psurfDst, ULONG iSolidColor, LONG lNumRects, RECTL* pRects);
INT		PixmapFill(PMirrSurf psurfDst, INT left, INT top, INT width, INT height,
				   LONG lNumRects, RECTL* pRects);
BOOL	Glyph(PDEV* ppdev, SURFOBJ* psoDst, RECTL* pRect, ULONG ulForeColor);

BOOL	CreateDeviceBitmap(PPDEV ppdev, PMirrSurf psurf);
BOOL	DestroyDeviceBitmap(PPDEV ppdev, int id);
BOOL	TimeStamp(PPDEV ppdev, int type, char* str);


/* refer to \mirror\disp\driver.h@line: 80 */

//
/* refer to \mirror\disp\driver.h@line: 83 */
/* refer to \mirror\disp\driver.h@line: 84 */
//

/* refer to \mirror\disp\driver.h@line: 87 */

/* refer to \mirror\disp\driver.h@line: 89 */   // Name of the DLL in UNICODE
/* refer to \mirror\disp\driver.h@line: 90 */  // All debug output is prefixed
/* refer to \mirror\disp\driver.h@line: 91 */        // Four byte tag (characters in
/* refer to \mirror\disp\driver.h@line: 92 */
/* refer to \mirror\disp\driver.h@line: 93 */

#include "thincDrvList.h"
#include "thincCmdBlock.h"
#include "thincCmdQue.h"
#include "common.h"

#endif

