/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/******************************Module*Header*******************************\
*
*                          *********************
*                          * THINC Driver CODE *
*                          *********************
*
* Module Name: other.c
*
* This module contains the functions drawing lines, curves and other advanced
* operations.
*
***************************************************************************/

#define STROKE_DBG_LEV 2

#include "driver.h"
#include "MirrCmdProto.h"
#include "clip.h"

#define GRADIENT_DBG_LEV 2
#define ALPHABLEND_DBG_LEV 2
#define TRANSPARENT_DBG_LEV 2

BOOL
  DrvAlphaBlend(
    IN SURFOBJ  *psoDst,
    IN SURFOBJ  *psoSrc,
    IN CLIPOBJ  *pco,
    IN XLATEOBJ  *pxlo,
    IN RECTL  *prclDest,
    IN RECTL  *prclSrc,
    IN BLENDOBJ  *pBlendObj
    )
{
	PMirrSurf	psurfDst = (PMirrSurf)(psoDst->dhsurf);
	PMirrSurf	psurfSrc = (PMirrSurf)(psoSrc->dhsurf);
	PPDEV		ppdev = (PPDEV)(psoDst->dhpdev);
	BOOL		res;
	BOOL		doRawUpdate = FALSE;

	int			i;
	int			lNumRects;

	RECTL*		pRects;


#if DBG
	TimeStamp(ppdev, TIME_STAMP_BEFORE_DRAWING, __FUNCTION__);		
#endif

	if (ppdev == NULL)
	{
		ppdev = (PPDEV)(psoSrc->dhpdev);
	}
	//BlockOnRequest(psoDst, psoSrc);

	if (psurfDst == 0)
	{
		//
		// It is a DIB, not a device managed surface
		//
		goto punt;
	}

	//ASSERT((pco != NULL), " DrvAlphaBlend: Expect non-null pco");

	doRawUpdate = TRUE;

	vClip(pco,
		prclSrc,	//prclSrc, 
		prclDest, 
		NULL,
		&pRects,
		&lNumRects,
		psurfSrc,
		psurfDst);

punt:

	START_DRAWING;
	res =  EngAlphaBlend(
		psoDst,
		psoSrc,
		pco,
		pxlo,
		prclDest,
		prclSrc,
		pBlendObj
    );
	FINISH_DRAWING;


	if (doRawUpdate)
	{
		for (i = 0; i < lNumRects; i++)
		{
			RawUpdate(psoDst, &pRects[i]);
   		}
		vFreeRectList(&pRects);
	}

	return res;
}

BOOL
  DrvFillPath(
    IN SURFOBJ  *psoDst,
    IN PATHOBJ  *ppo,
    IN CLIPOBJ  *pco,
    IN BRUSHOBJ  *pbo,
    IN POINTL  *pptlBrushOrg,
    IN MIX  mix,
    IN FLONG  flOptions
    )
{
	return TRUE;
}

BOOL
  DrvGradientFill(
    IN SURFOBJ  *psoDst,
    IN CLIPOBJ  *pco,
    IN XLATEOBJ  *pxlo,
    IN TRIVERTEX  *pVertex,
    IN ULONG  nVertex,
    IN PVOID  pMesh,
    IN ULONG  nMesh,
    IN RECTL  *prclExtents,
    IN POINTL  *pptlDitherOrg,
    IN ULONG  ulMode
    )
{
	PMirrSurf	psurfDst = (PMirrSurf)(psoDst->dhsurf);
	PPDEV		ppdev = (PPDEV)(psoDst->dhpdev);
	BOOL		res;
	BOOL		doRawUpdate = FALSE;

	int			i;
	int			lNumRects = 0;

	RECTL*		pRects = NULL;

	if (psurfDst == 0)
	{
		//
		// It is a DIB, not a device managed surface
		//
		goto punt;
	}

#if DBG
	TimeStamp(ppdev, TIME_STAMP_BEFORE_DRAWING, __FUNCTION__);		
#endif
	//ASSERT((pco != NULL), " DrvGradientFill: Expect non-null pco");

	doRawUpdate = TRUE;

	vClip(pco,
		NULL,	//prclSrc, 
		prclExtents, 
		NULL,
		&pRects,
		&lNumRects,
		NULL,
		psurfDst);

punt:

	START_DRAWING;
	res = EngGradientFill(
				psoDst,
				pco,
				pxlo,
				pVertex,
				nVertex,
				pMesh,
				nMesh,
				prclExtents,
				pptlDitherOrg,
				ulMode
				);
	FINISH_DRAWING;

	if (doRawUpdate)
	{
		for (i = 0; i < lNumRects; i++)
		{
			RawUpdate(psoDst, &pRects[i]);
   		}
		vFreeRectList(&pRects);
	}

	return res;
}

BOOL
  DrvLineTo(
    SURFOBJ  *psoDst,
    CLIPOBJ  *pco,
    BRUSHOBJ  *pbo,
    LONG  x1,
    LONG  y1,
    LONG  x2,
    LONG  y2,
    RECTL  *prclBounds,
    MIX  mix
    )
{
	PMirrSurf	psurfDst = (PMirrSurf)(psoDst->dhsurf);
	PPDEV		ppdev = (PPDEV)(psoDst->dhpdev);
	BOOL		res;
	BOOL		doRawUpdate = FALSE;

	int			i;
	int			lNumRects = 0;

	RECTL*		pRects = NULL;

	if (psurfDst == 0)
	{
		//
		// It is a DIB, not a device managed surface
		//
		goto punt;
	}


#if DBG
	TimeStamp(ppdev, TIME_STAMP_BEFORE_DRAWING, __FUNCTION__);		
#endif

	doRawUpdate = TRUE;

	vClip(pco,
		NULL,	//prclSrc, 
		NULL, 
		NULL,
		&pRects,
		&lNumRects,
		NULL,
		psurfDst);
punt:
	START_DRAWING;
	EngLineTo(
		psoDst,
		pco,
		pbo,
		x1,
		y1,
		x2,
		y2,
		prclBounds,
		mix);

	FINISH_DRAWING;
	if (doRawUpdate)
	{
		if (lNumRects > 0) 
		{
			for (i = 0; i < lNumRects; i++)
			{
				RECTL rct = pRects[i];
				if ((rct.bottom - rct.top == 1) || (rct.right - rct.left))
				{
					PMirrSurf psurfDst = (PMirrSurf)(psoDst->dhsurf);
					INT       iSolidColor = *((INT*) ((char*)(psurfDst->pscan) + 
						((psurfDst->lDelta * rct.top) + rct.left * (psurfDst->ulBitCount >> 3))));
					SolidFill(psurfDst, iSolidColor, 1, &rct);
				}
				else
				{
					RawUpdate(psoDst, &pRects[i]);
				}
			}
			vFreeRectList(&pRects);
		}
		else
		{
			if ((prclBounds->bottom - prclBounds->top == 1) || (prclBounds->right - prclBounds->left))
			{
				PMirrSurf psurfDst = (PMirrSurf)(psoDst->dhsurf);
				INT       iSolidColor = *((INT*) ((char*)(psurfDst->pscan) + 
					((psurfDst->lDelta * prclBounds->top) + prclBounds->left * (psurfDst->ulBitCount >> 3))));
				SolidFill(psurfDst, iSolidColor, 1, prclBounds);
			}
			else
			{
				RawUpdate(psoDst, prclBounds);		
			}
		}
	}
	return TRUE;
}

BOOL
  DrvPlgBlt(
    IN SURFOBJ  *psoDst,
    IN SURFOBJ  *psoSrc,
    IN SURFOBJ  *psoMsk,
    IN CLIPOBJ  *pco,
    IN XLATEOBJ  *pxlo,
    IN COLORADJUSTMENT  *pca,
    IN POINTL  *pptlBrushOrg,
    IN POINTFIX  *pptfx,
    IN RECTL  *prcl,
    IN POINTL  *pptl,
    IN ULONG  iMode
    )
{
	//BlockOnRequest(psoDst, psoSrc);

	return TRUE;
}

BOOL
  DrvStretchBlt(
    IN SURFOBJ  *psoDst,
    IN SURFOBJ  *psoSrc,
    IN SURFOBJ  *psoMask,
    IN CLIPOBJ  *pco,
    IN XLATEOBJ  *pxlo,
    IN COLORADJUSTMENT  *pca,
    IN POINTL  *pptlHTOrg,
    IN RECTL  *prclDest,
    IN RECTL  *prclSrc,
    IN POINTL  *pptlMask,
    IN ULONG  iMode
    )
{
	//BlockOnRequest(psoDst, psoSrc);
	PMirrSurf	psurfDst = (PMirrSurf)(psoDst->dhsurf);
	PMirrSurf	psurfSrc = (PMirrSurf)(psoSrc->dhsurf);
	PPDEV		ppdev = (PPDEV)(psoDst->dhpdev);
	BOOL		res;
	BOOL		doRawUpdate = FALSE;

	int			i;
	int			lNumRects = 0;

	RECTL*		pRects = NULL;

	if (psurfDst == 0)
	{
		//
		// It is a DIB, not a device managed surface
		//
		goto punt;
	}

#if DBG
	TimeStamp(ppdev, TIME_STAMP_BEFORE_DRAWING, __FUNCTION__);		
#endif

	//ASSERT((pco != NULL), " DrvStretchBlt: Expect non-null pco");

	doRawUpdate = TRUE;

	vClip(pco,
		prclSrc,	//prclSrc, 
		prclDest, 
		NULL,
		&pRects,
		&lNumRects,
		psurfSrc,
		psurfDst);

punt:
	START_DRAWING;
	EngStretchBlt(
		psoDst,
		psoSrc,
		psoMask,
		pco,
		pxlo,
		pca,
		pptlHTOrg,
		prclDest,
		prclSrc,
		pptlMask,
		iMode
    );

	FINISH_DRAWING;
	if (doRawUpdate)
	{
		for (i = 0; i < lNumRects; i++)
		{
			RawUpdate(psoDst, &pRects[i]);
   		}
	}
	vFreeRectList(&pRects);
	return TRUE;
}

BOOL
  DrvStretchBltROP(
    IN SURFOBJ  *psoDst,
    IN SURFOBJ  *psoSrc,
    IN SURFOBJ  *psoMask,
    IN CLIPOBJ  *pco,
    IN XLATEOBJ  *pxlo,
    IN COLORADJUSTMENT  *pca,
    IN POINTL  *pptlHTOrg,
    IN RECTL  *prclDest,
    IN RECTL  *prclSrc,
    IN POINTL  *pptlMask,
    IN ULONG  iMode,
    IN BRUSHOBJ  *pbo,
    IN DWORD  rop4
    )
{
	//BlockOnRequest(psoDst, psoSrc);

	return TRUE;
}

BOOL
  DrvStrokeAndFillPath(
    IN SURFOBJ  *psoDst,
    IN PATHOBJ  *ppo,
    IN CLIPOBJ  *pco,
    IN XFORMOBJ  *pxo,
    IN BRUSHOBJ  *pboStroke,
    IN LINEATTRS  *plineattrs,
    IN BRUSHOBJ  *pboFill,
    IN POINTL  *pptlBrushOrg,
    IN MIX  mixFill,
    IN FLONG  flOptions
    )
{
	//BlockOnRequest(psoDst, NULL);

	return TRUE;
}

BOOL
  DrvTransparentBlt(
    IN SURFOBJ  *psoDst,
    IN SURFOBJ  *psoSrc,
    IN CLIPOBJ  *pco,
    IN XLATEOBJ  *pxlo,
    IN RECTL  *prclDst,
    IN RECTL  *prclSrc,
    IN ULONG  iTransColor,
    IN ULONG  ulReserved
    )
{
	PMirrSurf	psurfDst = (PMirrSurf)(psoDst->dhsurf);
	PMirrSurf	psurfSrc = (PMirrSurf)(psoSrc->dhsurf);
	PPDEV		ppdev = (PPDEV)(psoDst->dhpdev);
	BOOL		res;
	BOOL		doRawUpdate = FALSE;

	int			i;
	int			lNumRects = 0;

	RECTL*		pRects = NULL;

	if (psurfDst == 0)
	{
		//
		// It is a DIB, not a device managed surface
		//
		goto punt;
	}


#if DBG
	TimeStamp(ppdev, TIME_STAMP_BEFORE_DRAWING, __FUNCTION__);		
#endif

	doRawUpdate = TRUE;

	vClip(pco,
		prclSrc,	//prclSrc, 
		prclDst, 
		NULL,
		&pRects,
		&lNumRects,
		psurfSrc,
		psurfDst);

punt:

	START_DRAWING;

	res =   EngTransparentBlt(
				psoDst,
				psoSrc,
				pco,
				pxlo,
				prclDst,
				prclSrc,
				iTransColor,
				ulReserved);
	FINISH_DRAWING;

	if (doRawUpdate)
	{
		if (lNumRects != 0)
		{
			for (i = 0; i < lNumRects; i++)
			{
				RawUpdate(psoDst, &pRects[i]);
			}
			vFreeRectList(&pRects);
		}
		else
		{
			RawUpdate(psoDst, prclDst);
		}
	}
	return res;
}
