/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/******************************Module*Header*******************************\
*
*                          *********************
*                          * THINC Driver CODE *
*                          *********************
*
* Module Name: pointer.c
*
* This module contains the functions for mouse cursors.
*
***************************************************************************/

//#define DBG 1
#define PTR_DBG_LEV 2
#include "driver.h"
#include "MirrCmdProto.h"

VOID
  DrvMovePointer(
    IN SURFOBJ  *pso,
    IN LONG  x,
    IN LONG  y,
    IN RECTL  *prcl
    )
{
	PPDEV	ppdev = PPDEV_OF_PSURFOBJ(pso);

	//DISPDBG((0, " DrvMovePointer is called\n"));
	if (ppdev->drvManagedCursor)
	{
		//
		// We don't want the cursor to appear on the surface.
		//
		return;
	}
	//
	// We are unable to deal with the current cursor. Punt it back 
	// to GDI Engine. It will draw it on the surface.
	//
	//EngMovePointer(
	//	pso,
	//	x,
	//	y,
	//	prcl);
}

ULONG
  DrvSetPointerShape(
    IN SURFOBJ  *psoDst,
    IN SURFOBJ  *psoMask,
    IN SURFOBJ  *psoColor,
    IN XLATEOBJ  *pxlo,
    IN LONG  xHot,
    IN LONG  yHot,
    IN LONG  x,
    IN LONG  y,
    IN RECTL  *prcl,
    IN FLONG  fl
    )

{
	
	PPDEV		ppdev = PPDEV_OF_PSURFOBJ(psoDst);
	PMirrSurf	psurfDst = (PMirrSurf)(psoDst->dhsurf);
	ULONG		res;
	INT			cx = psoMask->sizlBitmap.cx;
	INT			cy = psoMask->sizlBitmap.cy;
	PCHAR		mask;
	PCHAR		pvScan;
	LONG		delat;
	LONG		stride;
	PThincCmd	pCmd;
	
	//DISPDBG((0, " DrvSetPointerShape is called\n"));

	if (psoColor != NULL)
	{
		//
		// We are unable to deal with cursor with color
		//
		ppdev->drvManagedCursor = FALSE;
		return SPS_DECLINE;
		//goto punt;
	}

	if (ppdev->drvManagedCursor)
	{
		//DISPDBG((0, " DrvSetPointerShape: Free the memory for previous cursor mask\n"));
		EngFreeMem(ppdev->cursorMask);
	}


	stride = (cx + 7) >> 3;
	
	mask = (PCHAR)thincAlloc(stride * cy, "STPS");

	if (mask != NULL)
	{
		PVOID	data;
		INT		datasize;
		INT		delta;
		INT		i, offset;

		//DISPDBG((0, " DrvSetPointerShape: New memory block for cursor mask was successfully allocated\n"));

		ppdev->drvManagedCursor = TRUE;
		ppdev->xHot = xHot;
		ppdev->yHot = yHot;
		ppdev->cursorHeight = cx;
		ppdev->cursorWidth = cy >> 1;
		ppdev->cursorMask = mask;
		ppdev->cursorSize = stride * cy;

		pvScan = psoMask->pvScan0;
		delta = psoMask->lDelta;

		//DISPDBG((0, " DrvSetPointerShape: cx=%d, cy=%d, stride=%d, delta=%d\n",
		//	cx, cy, stride, delta));

		for (i = 0; i < cy; i++)
		{
			memcpy(mask, pvScan, stride);
			mask += stride;
			pvScan += delta;
		}

        //
		// Add a cursor change command.
		//
	
		datasize = THINC_HEADERSIZE + T_CURSOR_CHANGE_SIZE + ppdev->cursorSize;

		pCmd = thincCmdBlockAlloc(&(ppdev->pCmdBlock), datasize, &offset);

		if (pCmd != NULL)
		{
			PVOID data = pCmd->cmdData;
			PVOID cursormask = (PCHAR)data + THINC_HEADERSIZE + T_CURSOR_CHANGE_SIZE;
			MirrCmdHeader* header = (MirrCmdHeader*)data;
			thinc_cursorChange* cursorchange =(thinc_cursorChange*)((PCHAR)data + THINC_HEADERSIZE);

			header->type = T_CURSOR_CHANGE;
			header->dstId = psurfDst->id;
			header->srcId = NULL_SURF_ID;

			cursorchange->xhot = xHot;
			cursorchange->yhot = yHot;
			memcpy(cursormask, ppdev->cursorMask, ppdev->cursorSize);
#ifdef TIME_STAMP
			// Time stamp the command
			ppdev->pfQuerySystemTime(&pCmd->timeStamp[0]);
#endif
			if (!thincPushCmdWithOffset(ppdev, ppdev->pCmdBlock, offset))
			{
				DISPDBG((PTR_DBG_LEV, "  DrvSetPointerShape -- pushing command failed\n")); 
				thincFree(data);
			}					   
		}
	}
	else
	{
		//
		// Fail to allocate the memory for the mask.
		// 
		ppdev->drvManagedCursor = FALSE;
		return SPS_DECLINE;
		//goto punt;
	}

	//DISPDBG((0, " DrvSetPointerShape: DrvManagedCursor:%d, size:%d, height
//punt:
//  res = EngSetPointerShape(
//			pso,
//			psoMask,
//			psoColor,
//			pxlo,
//			xHot,
//			yHot,
//			x,
//			y,
//			prcl,
//			fl
//		);
//
//  return res;
	return SPS_ACCEPT_NOEXCLUDE;

}
