/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/******************************Module*Header*******************************\
*
*                       **************************
*                       * THINC Related Routines *
*                       **************************
*
* Module Name: thincDrvList.c
*
* This module contains the functions that are related with THINC 
*
\**************************************************************************/


#include "thincCmdQue.h"
#include "driver.h"


#define CQ_DBGLEV 10

#define containerPtr(structPtr, memberAddr, member)	\
	(structPtr)((char*)memberAddr - (int)(&(((NULL)(structPtr))->member)))

VOID thincInitCmdQueue(PThincCmdPtrQueue pThincCmdPtrQueue)
{
	int i;

	for (i = 0; i < pThincCmdPtrQueue->cmdQueueSize; i++)
	{
		pThincCmdPtrQueue->ThincCmdPtrList[i].offset = 0;
		pThincCmdPtrQueue->ThincCmdPtrList[i].pCmdBlock = NULL;
	}

	pThincCmdPtrQueue->cmdHeadIndex = 0;
	pThincCmdPtrQueue->cmdTailIndex = 0;

}

INT thincCreateCmdQueue(PThincCmdPtrQueue* ppThincCmdPtrQueue)
{
	ULONG tag = *((ULONG*)("CMDQ"));
	int size = MAX_CMD_QUEUE;
	PThincCmdPtrQueue pThincCmdPtrQueue = NULL;

	DISPDBG((0,"Initialize Display Command Queue\n"));

	while (size >= 0 && pThincCmdPtrQueue == NULL) 
	{
		// Allocate the space for pThincCmdPtrQueue. If failed, 
		// try a small size.
		pThincCmdPtrQueue = EngAllocMem(
			FL_NONPAGED_MEMORY,
			sizeof(ThincCmdPtr)*size + sizeof(ThincCmdPtrQueue),
			tag);
		if (pThincCmdPtrQueue != NULL)
		{
			pThincCmdPtrQueue->cmdQueueSize = size;
			pThincCmdPtrQueue->mask = size-1;
			*ppThincCmdPtrQueue = pThincCmdPtrQueue;
			thincInitCmdQueue(pThincCmdPtrQueue);
		}
		else 
		{
			size = size >> 1;
		}
	}
	DISPDBG((0,"Initialize Display Command Queue...Done, size = %d\n", pThincCmdPtrQueue->cmdQueueSize));
	return size;
}

BOOL pushCmd(PThincCmdPtrQueue pThincCmdPtrQue, ThincCmdPtr cmd)
{
	if (QUEUEFULL(pThincCmdPtrQue))
	{
		// The queue is full
		return FALSE;
	}
	else
	{	
		DISPDBG((CQ_DBGLEV, "head=%d, tail=%d\n", pThincCmdPtrQue->cmdHeadIndex, pThincCmdPtrQue->cmdTailIndex));
		pThincCmdPtrQue->ThincCmdPtrList[pThincCmdPtrQue->cmdTailIndex++] = cmd;
		DISPDBG((CQ_DBGLEV, "head=%d, tail=%d\n", pThincCmdPtrQue->cmdHeadIndex, pThincCmdPtrQue->cmdTailIndex));
		pThincCmdPtrQue->cmdTailIndex &= pThincCmdPtrQue->mask;
		DISPDBG((CQ_DBGLEV, "head=%d, tail=%d\n", pThincCmdPtrQue->cmdHeadIndex, pThincCmdPtrQue->cmdTailIndex));
	}
	return TRUE;
}

BOOL pushCmdWithOffset(PThincCmdPtrQueue pThincCmdPtrQue, PThincCmdBlock pCmdBlock, INT offset)
{
	PThincCmdPtr pThincCmdPtr;
	PPDEV		 ppdev;
	if (QUEUEFULL(pThincCmdPtrQue))
	{
		RIP("pushCmdWithOffset: Command Queue is full\n");
		// The queue is full
		return FALSE;
	}
	else
	{	
		DISPDBG((CQ_DBGLEV, "head=%d, tail=%d\n", pThincCmdPtrQue->cmdHeadIndex, pThincCmdPtrQue->cmdTailIndex));
		pThincCmdPtr = &(pThincCmdPtrQue->ThincCmdPtrList[pThincCmdPtrQue->cmdTailIndex++]);
		pThincCmdPtr->pCmdBlock = pCmdBlock;
		pThincCmdPtr->offset = offset;
		DISPDBG((CQ_DBGLEV, "head=%d, tail=%d\n", pThincCmdPtrQue->cmdHeadIndex, pThincCmdPtrQue->cmdTailIndex));
		pThincCmdPtrQue->cmdTailIndex &= pThincCmdPtrQue->mask;
		DISPDBG((CQ_DBGLEV, "head=%d, tail=%d\n", pThincCmdPtrQue->cmdHeadIndex, pThincCmdPtrQue->cmdTailIndex));

		//ppdev = containerPtr(PPDEV, pThincCmdPtrQue, pCmdPtrQueue);
		//if (ppdev->bCopyingCmds && pCmdBlock != NULL)
		//{
		//	//
		//	// Increase the command counter when this is not a seperator and command copying is going on
		//	//
		//	
		//}
	}
	return TRUE;
}

PThincCmdPtr popCmd(PThincCmdPtrQueue pThincCmdPtrQue)
{
	PThincCmdPtr ret = NULL;
	if (!QUEUEEMPTY(pThincCmdPtrQue))
	{
		ret = &(pThincCmdPtrQue->ThincCmdPtrList[pThincCmdPtrQue->cmdHeadIndex++]);
		pThincCmdPtrQue->cmdHeadIndex &= pThincCmdPtrQue->mask;
	}
	return ret;
}

BOOL queueFull(PThincCmdPtrQueue pThincCmdPtrQue)
{
	// Actually, the slot of ppThincCmdPtrQue[cmdHeadIndex-1] will never be used.
	// But don't worry, it is only a pointer. It doesn't waste too much space.
	return (pThincCmdPtrQue->cmdHeadIndex == ((pThincCmdPtrQue->cmdTailIndex+1) & pThincCmdPtrQue->mask));
}

//BOOL writeCmd(PThincCmdPtrQueue pThincCmdPtrQue)
//{
//	PThincCmdPtr cmd;
//	cmd = makeCmd(str, NULL, 0);
//	if (cmd == NULL) 
//		return FALSE;
//	DISPDBG((CQ_DBGLEV, "writeCmd()....Memory allocated\n"));
//    
//	if (!pushCmd(pThincCmdPtrQue, cmd))
//	{
//		EngFreeMem(cmd);
//		return FALSE;
//	}
//
//	return TRUE;
//}
//
//PThincCmdPtr retrieveCmd(PThincCmdPtrQueue pThincCmdPtrQue)
//{
//	PThincCmdPtr ret = NULL;
//	if (pThincCmdPtrQue->cmdHeadIndex != pThincCmdPtrQue->cmdTailIndex)
//	{
//		ret = pThincCmdPtrQue->ppThincCmdPtr[pThincCmdPtrQue->cmdHeadIndex];
//	}
//	return ret;
//}
//

BOOL thincPushCmdWithOffset(PPDEV ppdev, PThincCmdBlock pCmdBlock, INT offset)
{
	//if (ppdev->bCopyingCmds)
	//{
	//	ppdev->couterDrawCmds++;
	//}
	BOOL res = pushCmdWithOffset(ppdev->pCmdPtrQueue, pCmdBlock, offset);
#if 1
	if (res && ppdev->bServerOn /*&& ppdev->thincSignals.bReqNewCommandSignal*/)
	{
		ppdev->pfSetEvent(ppdev->pEventNewCmd);
		ppdev->thincSignals.bReqNewCommandSignal = FALSE;
	}
#endif
	return res;
}
