/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/******************************Module*Header*******************************\
*
*                       **************************
*                       * THINC Related Routines *
*                       **************************
*
* Module Name: thincDrvList.c
*
* This module contains the functions that are related with THINC 
*
***************************************************************************/


#include "thincDrvList.h"
#include "debug.h"
#include "ntddvdeo.h"
#include "utils.h"

PThincDrvNode thincDrvListCreate()
{
	PThincDrvNode pNode;

	pNode = EngAllocMem(FL_NONPAGED_MEMORY | FL_ZERO_MEMORY,
						sizeof(ThincDrvNode),
						MKTAG("LIST"));
	if (pNode != NULL)
	{
		pNode->element = NULL;
		pNode->next = pNode->prev = pNode;
	}
	return pNode;
}

PVOID thincDrvListReadFirstNode(PThincDrvNode pList)
{
	ASSERT((pList != NULL), "thincDrvListReadFirstNode: pList == NULL");
	return pList->next->element;
}

PVOID thincDrvListReadAndDeleteFirstNode(PThincDrvNode pList)
{
	PThincDrvNode pNode;
	PVOID element;

	ASSERT((pList != NULL), "thincDrvListReadAndDeleteOne: pList == NULL");

	pNode = pList->next;
	element = pNode->element;
	pList->next = pNode->next;
	pNode->next->prev = pList;
	
	if (element != NULL)
	{
		// it is not the header node
		EngFreeMem(pNode);
	}
	return element;
}

PThincDrvNode thincDrvListAdd(PThincDrvNode pList, PVOID data)
{          
	PThincDrvNode pNode;

	ASSERT((pList != NULL), "thincDrvListAdd: pList == NULL");

    pNode = thincAllocZero(sizeof(ThincDrvNode), "LIST");
	
	if (pNode == NULL)
	{
		return NULL;
	}

	pNode->element = data;
	pList->prev->next = pNode;
	pNode->prev = pList->prev;
	pNode->next = pList;
	pList->prev = pNode;

	return pNode;
}

BOOL thincDrvListDestroy(PThincDrvNode pList, FreeMem pfFreeMem)
{                    
	PThincDrvNode pNode;
	PThincDrvNode pNodeToFree;

	ASSERT((pList != NULL), "thincDrvListDestroy: pList == NULL");

	pNode = pList->next;
	
	while (pNode->next != pNode)
	{
		pNodeToFree = pNode;
		pNode = pNode->next;
		if (pNodeToFree->element != NULL && pfFreeMem != NULL)
		{
			pfFreeMem(pNodeToFree->element);
		}
		EngFreeMem(pNodeToFree);
	}

	//
	// Destroy the last node, which is usually the header node
	//
	if (pNodeToFree->element != NULL && pfFreeMem != NULL)
	{
		pfFreeMem(pNodeToFree->element);
	}

	EngFreeMem(pNodeToFree);

	return TRUE;
}

BOOL thincDrvListRemove(PThincDrvNode pNode)
{
	ASSERT((pNode != NULL), "thincDrvListRemove: pNode == NULL");
	if (pNode->next == pNode)
	{
		// It is the only node in the list and couldn't be removed.
		return FALSE;
	}
	pNode->next->prev = pNode->prev;
	pNode->prev->next = pNode->next;
	
	thincFree(pNode);
	
	return TRUE;
}
